package xmlhttp.framework;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.security.cert.X509Certificate;
import java.util.Enumeration;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Serializer;

import org.apache.log4j.Logger;
import xmlhttp.framework.messaging.HttpServletSOAPMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessage;
import za.co.solms.framework.ooxml.soap.processing.SOAPProcessor;

/** Controller to receive SOAP messages, which is handed to the SOAP OOXML Framework.
 * This servlet serves as the basis for a generic HTTP layer for the web services
 * framework, but it is currently quite basic, as well as tied to XMLVend.
 * 
 * @author dawidl
 * @version 0.1 (10 Oct 2005)
 */
public class MessageReceiverServlet extends HttpServlet
{
	@Override
	public void init() throws ServletException
	{
		// XOM Parser
		xomBuilder = new Builder();
		
		// SOAP Engine
		try
		{
			soapEngine = (SOAPProcessor)getServletContext().getAttribute("soap.engine");
			if (soapEngine == null)
			{
				throw new Exception("No SOAPProcessor in context");
			}
		}
		catch (Exception e)
		{
			throw new ServletException("Failed to get reference to SOAP Engine", e);
		}
		
		// Get WSDL Document
		try
		{
			logger.debug("Obtaining WSDL from SOAP Engine");
			wsdlDoc = xomBuilder.build( soapEngine.getWSDLLocator().getWSDLLocation().openStream() );
			// TODO: Fix WSLD URL to reflect accurate URL of this server
			logger.warn("WSDL URL is not fixed to reflect correct URL of this server (in this version)");
		}
		catch (Exception e)
		{
			logger.error("Unable to read WSDL (and thus unable to send to client)", e);
		}
	}
	
	
	@Override
	protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
	{
		try
		{
			// TODO Initial checks to see whether we should even try to parse
			//...
			
			
			// Get SOAP Message from client input stream
			logger.debug("Parsing XML Message");
			Document d = null;
			try
			{
				d = xomBuilder.build(request.getInputStream());
			} 
			catch ( Exception e)
			{
				// If parsing fails, tell client
				response.sendError(HttpServletResponse.SC_BAD_REQUEST );
				return;
			}
			
			// Construct an HTTP SOAP Message with a handle to this request
			SOAPMessage reqMsg = new HttpServletSOAPMessage( request, d );
			
			// Hand SOAP Message to WS Framework
			logger.debug("Handing message to SOAP framework");
			SOAPMessage resMsg = soapEngine.process( reqMsg );
			
			// Send SOAP Response to Client
			// If we have a response, serialize it to the client
			if (resMsg != null && resMsg.getSOAPEnvelope() != null)
			{
				logger.debug("Sending response message to client");
				String xml = resMsg.getSOAPEnvelope().toXML();
				byte[] responseData = xml.getBytes();
				logger.debug("Response size: " + responseData.length);
				
				// Setup headers
				response.setContentType("text/xml");
				
				/* TODO The 3rd party GZIP response filter we are currently using 
				 * does not properly handle content-length (probably for reasons of 
				 * streaming, as it would otherwise have to cache in-memory like
				 * our in-house Request filter. So, for now, we do not specify
				 * content length explicitly (a better filter should be used to
				 * calculate this).
				 */
				//response.setContentLength( xml.length() );
				
				if (resMsg.isFault())
				{
					response.setStatus( HttpServletResponse.SC_INTERNAL_SERVER_ERROR );
				}
				
				//Comment out to simulate a server connection failure 
				//delay every alternate connection				
				/*
				if (timeout == true)
				{
					timeout = false;
					logger.info("Timing out client!!!");
					try
					{
						//delay response for 100sec
						Thread.sleep(30000);
					}catch(InterruptedException ie){}
					logger.info("Delay complete, sending response to client!");
				}
				else
				{
				 timeout = true;
				} 
				*/
				//Delay ends
				
				// Send Message
				response.getOutputStream().write( responseData );
				response.flushBuffer();
			}
			else
			{
				logger.debug("No response message");
			}
			
		} catch (Exception e)
		{
			throw new ServletException(e);
		}
		
	}
	
	@Override
	protected void doGet(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException
	{
		if (req.getParameter("wsdl") != null)
		{
			logger.debug("Sending WSDL");
			// Send back WSDL
			// TODO: Configure this in web.xml rather
			res.setContentType("text/xml");
			
			logger.debug("Using XOM serializer to send document");
			ByteArrayOutputStream bao = new ByteArrayOutputStream( 1024 );
			Serializer xomSerialiser = new Serializer( bao );
			xomSerialiser.write( wsdlDoc );
			res.getOutputStream().write( bao.toByteArray() );
		}
		else
		{
			// Default (not supported) behaviour
			super.doGet(req, res);
		}
	}
	

	private static final long serialVersionUID = 5733420687126585244L;
	private SOAPProcessor soapEngine;
	private Logger logger = Logger.getLogger( MessageReceiverServlet.class );
	private Builder xomBuilder;
	private Document wsdlDoc;
	private boolean timeout = false;	
}