package xmlhttp.framework.config;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import za.co.eskom.nrs.xmlvend.server.BaseRequestProcessor;
import za.co.eskom.nrs.xmlvend.server.faults.XMlVendFaultFactory;
import za.co.solms.framework.impl.ooxml.basic.types.FileSystemSchemaLocator;
import za.co.solms.framework.ooxml.soap.processing.SOAPProcessor;
import za.co.solms.framework.ooxml.soap.processing.WSDLLocator;

/** Sets up the OOXML Framework and processor(s). An instance of SOAPProcessor
 * is bound to the servlet context as "soap.engine" at context startup. */
public class FrameworkSetupListener implements ServletContextListener
{
	public FrameworkSetupListener(){}

	public void contextInitialized(ServletContextEvent evt)
	{
		// Setup Logging
		BasicConfigurator.configure();
		
		// SOAP Engine
		try
		{
			// Setup and unpack schemas if necessary
			setupSchemaDir();
			
			// Setup SOAP engine + fault factory
			File schemaDir = getSchemaDir();
			soapEngine = SOAPProcessor.Factory.getInstance( 
				new FirstInDirectoryWSDLLocator( schemaDir ), 
				new FileSystemSchemaLocator( schemaDir ) );
			
			// Place in context
			evt.getServletContext().setAttribute(SOAP_ENGINE, soapEngine);
		}
		catch (Exception e)
		{
			logger.fatal("SOAP Engine failed", e);
		}
		
		// Plug in the XMLVend SOAP Test Suite Processor
		// TODO Read processor instantiations from an external file
		try
		{
			soapEngine.addProcessor( new BaseRequestProcessor() );
		} catch (Exception e)
		{
			logger.error("Unable to register XMLVend Processor",e);
		}
		
		// Put values in the context for the benefit of e.g. JSP
		try
		{
			evt.getServletContext().setAttribute("appSupportDir", FrameworkSetup.getUserDirectory().toString() + "/xmlvend/test-server/" );
		} catch (IOException e)
		{
			logger.error( e );
		}
		
		// Faults are located anywhere recursively in the same place normal
		// responses are stored, as per XMLVend BaseRequestProcessor
		try
		{
			soapEngine.setFaultFactory( 
				new XMlVendFaultFactory( BaseRequestProcessor.getInstanceFileDir() ) );
		}
		catch (Exception e)
		{
			logger.error("Unable to configure custom fault factory", e);
		}
	}

	public void contextDestroyed(ServletContextEvent evt){}
	
	
	/** Returns the base directory of the schema catalog */
	public static File getSchemaDir() throws IOException
	{
		// FIXME Remove hard-coded reference to XMLVend
		return FrameworkSetup.getUserDirectory("xmlvend/test-server/schemas");
		//return ApplicationSupportUtility.getAppSupportDirectory("xmlvend/test-server/schemas");
	}
	
	/** Makes sure the schema directory contains schemas */
	public void setupSchemaDir() throws IOException
	{
		if ( getSchemaDir().list().length == 0)
		{
			logger.debug("No files in schema directory, unpacking default set");
			// Unzip default set of schemas
			// TODO Move to eskom-specific directory
			ZipInflator.inflateZipStream( 
				getClass().getClassLoader().getResourceAsStream("xmlvend-schemas-default.zip"),
				getSchemaDir() );
		}
	}
	
	
	private SOAPProcessor soapEngine;
	Logger logger = Logger.getLogger(FrameworkSetupListener.class);
	public static final String SOAP_ENGINE = "soap.engine";
	
	
	/** Dummy (useless) WSDL Locator - we're not currently using it */
	private class DummyWSDLSource implements WSDLLocator
	{
		public URL getWSDLLocation()
		{
			return null;
		}
	}
	
	/** A simple locator that always produces the first found WSDL in a given
	 * directory. Does *not* recurse.
	 */
	private class FirstInDirectoryWSDLLocator implements WSDLLocator
	{
		public FirstInDirectoryWSDLLocator( File directory )
		{
			if (directory.isDirectory() && directory.canRead() )
			{
				for (File f : directory.listFiles())
				{
					try
					{
						// Get first .wsdl file
						if (f.isFile() && f.getName().endsWith(".wsdl"))
						{
							l = f.toURL();
							break;
						}
					} catch (MalformedURLException e)
					{
						logger.error("Invalid WSDL file URL", e);
					}
				}
				logger.debug("WSDL Locator found WSDL: " + l);
			}
		}
		
		public URL getWSDLLocation()
		{
			return l;
		}
		
		private URL l;
	}
}
