package xmlhttp.framework.io;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;
import java.util.zip.GZIPInputStream;
import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletRequestWrapper;

import org.apache.log4j.Logger;


/** Upon creation, this Request Wrapper will read and cache all available HTTP body data from
 * the client. This is then decompressed (using the gzip format) to a second in-memory
 * data cache. All subsequently read data (e.g. from the getInputStream() service) will read
 * from this buffer.
 * 
 * In addition, this wrapper attempts to hide all traces of compression by fixing the
 * content-length header, and masking (hiding) any headers which give away compression,
 * such as 'Content-Encoding'.
 * 
 * Request compression is a standard HTTP 1.1 feature, although most web containers 
 * do not have the capability to do this themselves. This is in contrast to HTTP 
 * RESPONSE compression, which most containers do support.
 * 
 * @author Dawid Loubser (Solms TCD)
 * @version 0.1 (Sep 17, 2004)
 * @see http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.5
 */
public class HttpCompressedRequestWrapper extends HttpServletRequestWrapper
{
    // Our uncompressed request data (cache)
    private byte[] requestData;
    // Name of the content encoding header (case insensitive)
    public static final String HEADER_CONTENT_ENCODING = "Content-Encoding";
    // Logging
    private Logger logger = Logger.getLogger(HttpCompressedRequestWrapper.class);
    
    
    /** Constructor: Chain to an existing HTTPServletRequest.
     * @param request The original HTTP request.
     */
    public HttpCompressedRequestWrapper(HttpServletRequest request) throws IOException
    {
        super(request);
        
        // Cache (pre-read entirely) original (compressed) data, and uncompress
        byte[] reqData = readIntoBuffer(request);        
        this.requestData = expand(reqData);
    }
    
    
    /** 
     * @see javax.servlet.ServletRequestWrapper#getInputStream()
     */
    public ServletInputStream getInputStream() throws IOException
    {
        // Create a new in-memory input stream
        return new InMemoryServletInputStream(this.requestData);
    }
    
    
    /** 
     * @see javax.servlet.ServletRequestWrapper#getReader()
     */
    public BufferedReader getReader() throws IOException
    {
        // Create a new reader
        return new BufferedReader( new InputStreamReader( new ByteArrayInputStream( this.requestData)));
    }
    
    
    /** Overrides original (compressed) content length, now reports
     * the true (uncompressed) length.
     *  
     * @see javax.servlet.ServletRequestWrapper#getContentLength()
     */
    public int getContentLength()
    {
        return requestData.length;
    }
    
    
    /** 
     * Returns the specified request header. Modified, so that non content-encoding
     * is returned for components, i.e. remove all traces of compression.
     * 
     * @see javax.servlet.http.HttpServletRequestWrapper#getHeader(java.lang.String)
     */
    public String getHeader(String name)
    {
        if (HEADER_CONTENT_ENCODING.equalsIgnoreCase(name))
        {
            return null;
        }
        else
        {
            return super.getHeader(name);
        }
    }
    
    /** 
     * Returns all headers except anything which belies the original Content-Encoding:
     * 
     * @see javax.servlet.http.HttpServletRequestWrapper#getHeaderNames()
     */
    public Enumeration getHeaderNames()
    {
        // Cycle through original headers, only add the right ones to a new
        // vector, and return those as a new Enumeration.
        Enumeration originalHeaders = super.getHeaderNames();
        Vector moddedHeaders = new Vector();
        while (originalHeaders.hasMoreElements())
        {
            try
            {
                String hdr = (String)originalHeaders.nextElement();
                if ( !hdr.equalsIgnoreCase(HEADER_CONTENT_ENCODING) )
                {
                    moddedHeaders.add(hdr);
                }
            } catch (ClassCastException ignored)
            {
                // Shame...
            }
        }        
        return moddedHeaders.elements();
    }    
   
    
    /** Fills the internal buffer with all data that can be read from the HTTP
     * request: This effectively kills the streaming behaviour (i.e. we are caching
     * it) but is required for correct computation of, for instance, content-length.
     * @return the amount of data read from the client.
     */
    private byte[] readIntoBuffer(HttpServletRequest request) throws IOException
    {
        // Create a buffer just large enough to hold all data
        byte[] reqData = new byte[ super.getContentLength() ];
        
        // Our underlying request stream
        InputStream in = super.getInputStream();
        
        // Read data in byte-array chunks
        int bytesRead = 0;
        int startIndex = 0;
        int amtToRead = reqData.length;
        int totalBytesRead = 0;
        try
        {
	        do
	        {
	            // Read chunk
	            bytesRead = in.read(reqData, startIndex, amtToRead);
	            // Calculate ranges for next iteration
	            startIndex = bytesRead;
	            amtToRead = reqData.length - bytesRead;
	            // Increment total amout of data read
	            totalBytesRead += (bytesRead == -1) ? 0 : bytesRead;
	        }
	        while (bytesRead != -1);
	        
        } catch (EOFException ignored){}
        
        // Test if content-length header matches actual data read
        if (super.getContentLength() != totalBytesRead)
        {
            throw new IOException("Client 'Content-Length:' header does not match actual amount of data.");
        }
        
        return reqData;
    }

    
    /** Decompresses the given block of compressed data, producing
     * a new byte array, and returning it.
     * @param compressedData GZIP-compressed HTTP body data
     * @return a byte array of uncompressed data, almost always larger than the original data.
     */
    private byte[] expand(byte[] compressedData) throws IOException
    {
        // Chain a gzip input stream to the data from the byte array
        ByteArrayInputStream bis = new ByteArrayInputStream( compressedData );
        GZIPInputStream gis = new GZIPInputStream(bis, compressedData.length);
        
        // Copy to an array (data sink), uncompressing in the process
        // This is not very fast, can be changed later to do buffering etc.
        ByteArrayOutputStream bos = new ByteArrayOutputStream(compressedData.length);
        
        int read;
        do
        {
            read = gis.read();
            if (read != -1)
                bos.write(read);
        }
        while (read != -1);       
        
        byte[] uncompressedData = bos.toByteArray();
        
        // Debugging message
        int cLen = compressedData.length;
        int uLen = uncompressedData.length;
        double s = (double)uLen / (double)cLen * 100.0d;
        logger.info("Inflated request: " + cLen + " bytes -> " + uLen + " bytes (" + s + " %)");
        //System.out.println("[Server] Request: "+compressedData.length+" bytes inflated to "+uncompressedData.length+" bytes.");
        
        return uncompressedData;
    }    
}