package xmlhttp.framework.io;

import java.io.IOException;
import javax.servlet.*;
import javax.servlet.Filter;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * This filter enables compression as per the HTTP 1.1 standard, though only using the
 * common gzip encoding. It currently only caters for the rare case
 * of HTTP REQUEST compression, as standard Tomcat configuration takes care of the
 * HTTP RESPONSE.
 * 
 * Note: This filter should probably be first in your chain, as nothing else will understand
 * compressed data!
 * 
 * @author Dawid Loubser (Solms TCD)
 */
public class HttpCompressionFilter implements Filter
{

    /**
     * Default constructor. 
     */
    public HttpCompressionFilter()
    {
    }

    
    /** Init (does nothing)
     */
    public void init(FilterConfig cfg) throws ServletException{}

    
    /** Filter: Check if the HTTP request is encoded with gzip, and if so, chain the appropriate
     * request wrapper that will perform de-compression. It is not clear what happens if the web server
     * already uncompresses this (i.e. double-uncompression) - testing as required, as soon as we find
     * a web container that actually *does* de-compress HTTP requests.
     */
    public void doFilter(ServletRequest request, ServletResponse response,
            FilterChain chain) throws IOException, ServletException
    {
        // HANDLE REQUEST
        try
        {
            HttpServletRequest req = (HttpServletRequest)request;
            String encoding = req.getHeader("Content-Encoding");
            
            if (encoding == null || (encoding = encoding.trim()).equals(""))
            {
                // Do nothing (No Compression)
            }
            else if ("gzip".equalsIgnoreCase(encoding))
            {
                // Wrap in our decompression wrapper
                System.out.println("[Server] Client sent compressed request, de-compressing...");
                req = new HttpCompressedRequestWrapper(req);
            }
            else
            {
                // If other encoding, send 415 (Unsupported Media Type)
                ((HttpServletResponse)response).sendError(HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE, "Unsupported Media Type ("+encoding+")");
                // Not sure if this is the correct course of action...
                throw new IOException("CompressionFilter: Client sent unsupported content-encoding type ("+encoding+")");
            }
            
            // HANDLE RESPONSE
            // (Not done in this version, rely on web container (e.g. Tomcat: Setting in /conf/server.xml)
        
            // Invoke next object on chain
            chain.doFilter(req, response);
        }
        catch (ClassCastException e)
        {
            // In case the unthinkable happens
            throw new ServletException("CompressionFilter: Cast Failed (CRITICAL)",e);
        }
    }

    
    /** Called when this filter will be destroyed.
     */
    public void destroy() {}

}