package xmlhttp.framework.parsing;
import java.io.IOException;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import nu.xom.Builder;
import nu.xom.Element;
import nu.xom.ParsingException;

/** This filter automatically constructs a Document Object Model (XOM) of any XML
 * document that is the body of a POST request. It then stores this XOM (Element) as a request 
 * attribute OM_KEY ("xml.om"), making it easily available to any other components that 
 * form part of the request processing chain.
 * 
 * After it in the chain (in the case of web services) would typically be a filter to extract the
 * SOAP Parts.
 * 
 * TODO Testing is required to see if a single instance of XOM Builder can handle concurrency
 * 
 * @author dawidl
 * @version 0.1 (Sep 30, 2005)
 */
public class OMBuilderFilter implements Filter
{

	public void init(FilterConfig cfg) throws ServletException
	{
		this.config = cfg;
		xomBuilder = new Builder();
	}

	
	/** Performs filtering on the given request / response. Tries to determine
	 * if the request contains XML information, and if it does, constructs a DOM an
	 * places it in the request (as "dom").
	 *  
	 * @see javax.servlet.Filter#doFilter(javax.servlet.ServletRequest, javax.servlet.ServletResponse, javax.servlet.FilterChain)
	 */
	public void doFilter(ServletRequest req, ServletResponse res,
			FilterChain chain) throws IOException, ServletException
	{
		// We're assuming an HTTP environment
		HttpServletRequest request = (HttpServletRequest)req;
		
		// Only do post requests
		if ( request.getMethod().equalsIgnoreCase("POST") )
		{
			// Original idea was to check if the content type is reported to be XML (text/xml)
			// but that's not reliable, i.e. several mime types exist for XML, and of course,
			// people often don't adhere to proper MIME naming standards.

			// Build XOM and place in request
			try
			{
				Element om = xomBuilder.build( request.getInputStream() ).getRootElement();
				config.getServletContext().log(LP+"Request contains XML message (" + om.getValue().length() + " chars)");
				request.setAttribute(OM_KEY, om);
			} catch (ParsingException e)
			{
				// This means that no XML could be parsed from the request (go on)
			}
		}
		
		// And (drum roll) the show goes on...
	    chain.doFilter(req, res);
	}

	
	public void destroy()
	{
		// Do nothing
	}
	
	
	/** Used to bind the Object Model to the HTTP Request scope */
	public static final String OM_KEY = "xml.om";
	private FilterConfig config;
	private Builder xomBuilder;
	private static final String LP = "OMBuilderFilter: "; // Log Prefix
}