package xmlhttp.framework.parsing;

/** Expects a SOAP message as a XOM model in the request ("xml.om"). Extracts the
 * various SOAP parts, and places these as XOM models in the request:
 * 
 * "xml.om.soap.body"   - Main Body, as an Element (xom)
 * "xml.om.soap.header" - Header elements, as a Nodes (xom)
 * 
 * Extraction is done via simple XPath. The SOAP namespace 
 * is assumed to be: http://schemas.xmlsoap.org/soap/envelope/ 
 * 
 * @author dawidl
 */
import java.io.IOException;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import nu.xom.Element;
import nu.xom.Nodes;
import nu.xom.XPathContext;

public class SOAPExtractorFilter implements Filter
{

	public void init(FilterConfig cfg) throws ServletException
	{
		this.config = cfg;
		// Setup XPath Context
		xpContext = new XPathContext("soap",SOAP_NAMESPACE);
	}

	public void doFilter(ServletRequest req, ServletResponse res,
			FilterChain chain) throws IOException, ServletException
	{
		HttpServletRequest request = (HttpServletRequest)req;
		HttpServletResponse response = (HttpServletResponse)res;
		
		// Get Object Model
		Object om = request.getAttribute( OMBuilderFilter.OM_KEY );
		// Is it a XOM Element?
		if (om != null && om instanceof Element)
		{
			Element soap = (Element)om;
			
			// Header parts
			Nodes headerParts = soap.query("/soap:Envelope/soap:Header/*",xpContext);
			if (headerParts.size() > 0)
			{
				request.setAttribute(SOAP_HEADER_KEY, headerParts);
				config.getServletContext().log(LP+"SOAP header(s) read: " + headerParts.size());
			}
			
			// Body Part
			Nodes bodyParts = soap.query("/soap:Envelope/soap:Body/*",xpContext);
			if (bodyParts.size() == 1)
			{
				Element body = (Element)bodyParts.get(0);
				request.setAttribute(SOAP_BODY_KEY, body);
				config.getServletContext().log(LP+"SOAP Body read: " + body.getQualifiedName());
			}
			else if (bodyParts.size() > 1)
			{
				// TODO Send error response (illegal)
				config.getServletContext().log(LP+"Body contains more than one part (illegal)");
				response.sendError( HttpServletResponse.SC_BAD_REQUEST );
			}
		}

		// Continue
		chain.doFilter(request, response);
	}

	public void destroy()
	{
		// TODO Auto-generated method stub

	}

	private FilterConfig config;
	private XPathContext xpContext;
	private static final String LP = "SOAPExtractorFilter: "; // Log Prefix
	/** Used to store the Body XOM in request scope as an ELEMENT */
	public static final String SOAP_BODY_KEY = "xml.om.soap.body";
	/** Used to store the Headers XOM in request scope as a Nodes */
	public static final String SOAP_HEADER_KEY = "xml.om.soap.header";
	/** The assumed SOAP namespace (other variants deliberately not supported */
	public static final String SOAP_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/";
}
