package za.co.eskom.nrs.xmlvend.server;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;

import javax.servlet.http.HttpServletRequest;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.ParsingException;
import org.apache.log4j.Logger;
import xmlhttp.framework.config.FrameworkSetup;
import xmlhttp.framework.config.ZipInflator;
import xmlhttp.framework.messaging.HttpServletSOAPMessage;
import za.co.eskom.nrs.xmlvend.messages.ClientID;
import za.co.eskom.nrs.xmlvend.messages.ClientIDFactory;
import za.co.eskom.nrs.xmlvend.messages.MessageIdentifier;
import za.co.eskom.nrs.xmlvend.messages.MessageIdentifierFactory;
import za.co.eskom.nrs.xmlvend.messages.MessageUnidentifiableException;
import za.co.eskom.nrs.xmlvend.server.advice.LastResponseStore;
import za.co.eskom.nrs.xmlvend.server.advice.impl.InMemoryLastResponseStore;
import za.co.solms.framework.impl.ooxml.basic.types.MessageTypeExtractor;
import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.processing.XMLProcessor;
import za.co.solms.framework.ooxml.processing.XMLType;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import za.co.eskom.nrs.xmlvend.server.security.*;

/** This is the base processor for XMLVend Use Cases. For standard
 * requests, it locates a suitable response template (from annotated
 * file), is used to transform the request with (using the request doc as input),
 * and the response is returned. Non-XSLT (.xml) responses are also
 * accomodated, though sent back literally, with no transformation.
 * 
 * TODO: This class has not been optimised for performance. Extremely inefficient use
 * is made of XSLT, and conversion to and from XOM.
 * 
 * @author dawidl
 * @version 0.1 (Nov 3, 2005)
 */
public class BaseRequestProcessor implements XMLProcessor
{
	public BaseRequestProcessor()
	{
		// Setup XSLT Transformation
		// Java5 Built-in
		//System.setProperty("javax.xml.transform.TransformerFactory", "com.sun.org.apache.xalan.internal.xsltc.trax.TransformerFactoryImpl");
		// Saxon 8
		//System.setProperty("javax.xml.transform.TransformerFactory", "net.sf.saxon.TransformerFactoryImpl");
	}

	
	/** Initialisation */
	public void init(SchemaTypeRegistry typeRegistry) throws IOException
	{
		// Make sure there are responses (default setup)
		setupInstanceFileDir();
		
		// Construct Response Map
		try
		{
			responses = new ResponseMap( getInstanceFileDir(), typeRegistry );
		} catch (Exception e)
		{
			logger.fatal("Response message map failed", e);
		}
		
		// Setup store of last responses
		lastResponseStore = new InMemoryLastResponseStore();
		
		// Setup type extractor (to identify response messages)
		typeExtractor = new MessageTypeExtractor( typeRegistry );
	}
	
	
	/** For these messages, first store the response before sending to the client */
	@XMLType(name="CreditVendReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processCreditVend( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="CancelVendReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processCancelVend( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="EndBatchReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processEndBatch( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="StartBatchReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processStartBatch( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="PayAccReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processPayIssue( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="FreeIssueReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processFreeIssue( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="MeterCreditTransferReq", nameSpace=XMLVendConstants.XMLVEND_REVENUE_NAMESPACE)
	public XMLMessage processMeterCreditTransfer( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	@XMLType(name="CustReportFaultReq", nameSpace=XMLVendConstants.XMLVEND_METER_NAMESPACE)
	public XMLMessage processCustReportFault( XMLMessage input ) throws Exception
	{
		return processMessageWithStoredResponse( input );
	}
	
	@XMLType(name="ReversalAdviceReq", nameSpace=XMLVendConstants.XMLVEND_BASE_NAMESPACE)
	public XMLMessage getRevesalResponse( XMLMessage input ) throws ClientIDSSLException, Exception
	{
		logger.debug("Request received to reverse last response");
		
		//KPS 17/08/2006 - Check for SSL /TLS and client certificate CN same as clientID in message
		cnCheck.cnClientIDcheck(input);
		logger.debug("No client certificates or CN and ClientID match");
		
		// Check if message was processed (stored)
		XMLMessage lastResp = lastResponseStore.getLastResponseMessage( input );
		logger.debug("Message processed, reversing transaction");
		
		// Get response template
		return getResponseMessage( input );			 
	}
	
	/** For a 'LastRespAdviceReq' request, find the matching (stored) response, if
	 * any, and insert it in a correct response message to send to the client.
	 */
	@XMLType(name="LastRespAdviceReq", nameSpace=XMLVendConstants.XMLVEND_BASE_NAMESPACE)
	public XMLMessage getLastResponse( XMLMessage input ) throws ClientIDSSLException, Exception
	{
		//KPS 17/08/2006 - Check for SSL /TLS and client certificate CN same as clientID in message
		cnCheck.cnClientIDcheck(input);
		logger.debug("No client certificates or CN and ClientID match");

		logger.debug("Request received to retrieve last response");
		
		// Find last stored message
		XMLMessage lastResp = lastResponseStore.getLastResponseMessage( input );
		logger.debug("Last response found, constructing complete response message");
		
		// Get response template
		XMLMessage response = getResponseMessage( input );
		logger.debug("Response template found, inserting last response");
		
		// Insert last response into appropriate place in response message
		XMLMessage completeResponse = lastResponseAssembler.buildLastResponse( response, lastResp );
		return completeResponse;
	}
	
	
	/** Handles 'Base' requests, which means all sub-types will also be dispatched
	 * here unless overridden by another processing method 
	 * @see za.co.solms.framework.ooxml.processing.XMLProcessor#process(za.co.solms.framework.ooxml.XMLMessage)
	 */
	@XMLType(name="BaseReq", nameSpace=XMLVendConstants.XMLVEND_BASE_NAMESPACE)
	public XMLMessage process(XMLMessage input) throws Exception
	{
		logger.debug("Message received (type: " + input.getType() + ")");

		// SECURITY: Check if client used a certificate
		//HttpServletSOAPMessage message = (HttpServletSOAPMessage)input;
		/*
		Object cert = message.getHttpRequest().getAttribute("javax.servlet.request.X509Certificate");
		if (cert != null)
		{
			//X509Certificate x509 = (X509Certificate)cert;
			logger.debug("Client certificate: " + cert + " (" + cert.getClass().getName() + ")");
		}
		
		if (message.getHttpRequest().getUserPrincipal() != null)
		{
			logger.debug("User principal: " + message.getHttpRequest().getUserPrincipal());
		}*/


		//KPS 17/08/2006 - Check for SSL /TLS and client certificate CN same as clientID in message
		cnCheck.cnClientIDcheck(input);
		logger.debug("No client certificates or CN and ClientID match");
		
		return getResponseMessage(input);
	}
	

	/** Processes a request normally, and stores the response for later advice retrieval, etc */
	protected XMLMessage processMessageWithStoredResponse( XMLMessage input ) throws Exception
	{
		// Process as normal
		XMLMessage response = process( input );
		
		//KPS 13/10/2006 - check if request message ID match a already stored MsgID
		logger.debug("Checking MsgIDs");
		if (lastResponseStore.RequestMsgIDPresent( input ))
		{
			logger.debug("Message ID's match - This should not happen!!! An exception shoudl have been thrown.");			
		}
		else
		{
			logger.info("Request MsgID does not match stored response MsgID. Continuing to process request...");
		}
	
		// Store last response
		logger.debug("Storing response");
		lastResponseStore.storeMessage( response );
		
		// Return
		return response;
	}
	
	
	/** Locates and instrantiates the correct response message based on input type */
	protected XMLMessage getResponseMessage( XMLMessage input ) throws NoResponseFoundException, Exception
	{
		// Get Response Template
		File resp = responses.getResponse( input.getType() );
		
		if (resp != null)
		{
			String resName = resp.getName().toLowerCase();
			logger.debug("Responding with: " + resName);
			Document response = null;
			
			try
			{
			
				// Respond either with transformation, or literally (both supported)
				if (resName.endsWith(".xml"))
				{
					logger.warn("Response will be sent to client literally, and probably contains invalid values. Use XSLT instead.");
					response = respondLiterally( resp, input );
				}
				else if (resName.endsWith(".xsl"))
				{
					response = respondByTransformation( resp, input);
				}
			}
			catch (Exception e)
			{
				// Sending failed - log, and continue
				logger.error("Unable to send response", e);
				throw e;
			}
			
			// Wrap in XML Message
			XMLMessage responseMessage = new XMLMessage( response );
			
			// Assign type information (if known)
			try
			{
				SchemaType responseType = typeExtractor.deriveType( responseMessage.getMessage().getRootElement() );
				responseMessage.setType( responseType );
				logger.debug("Response message schema type: " + responseType);
				
			}
			catch (Exception e)
			{
				logger.warn("Unable to derive schema type of response message");
			}
			
			return responseMessage;
		}
		else
		{
			logger.debug("No response found for message, raising exception");
			//KPS - 27/07/2006 - Add exception message.
			throw new NoResponseFoundException("The requested Use Case is not supported");
		}
	}
	
	
	/** Create response message by transforming the input with the response */
	protected Document respondByTransformation( File responseTemplate, XMLMessage input) throws TransformerException, IOException, ParsingException
	{
		// Setup XSLT
		logger.debug("Creating XSL transformer");
		// TODO Optimise this (cache)
		Transformer transformer = TransformerFactory.newInstance().newTransformer( new StreamSource(responseTemplate) );

		
		// Setup input as a stream (this is very inefficient)
		byte[] xmlData = input.getMessage().toXML().getBytes();
		ByteArrayInputStream bis = new ByteArrayInputStream(xmlData);
		
		// Setup output as a stream
		ByteArrayOutputStream bos = new ByteArrayOutputStream(256);
		
		//set XSL respDateTime parameter - KPS - 3/04/2006
		//get response dateTime
		Date now = new Date();
		//setup the XML schema dateTime format
        SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'hh:mm:ss");
        //format the dateTime
        String respDateTime = formatter.format(now);
		//set XSL parameter
        transformer.setParameter("Date", respDateTime );
		
		// Transform input document using response
		logger.debug("Transforming request using response");
		transformer.transform( new StreamSource(bis), new StreamResult(bos));
		byte[] data = bos.toByteArray();
		
		logger.debug("Done (ready for sending to client)" + data.length + " bytes)");

		// Create XOM tree for result (also very inefficient!)
		return xomBuilder.build( new ByteArrayInputStream( data ) );
	}
	
	
	/** Create response message simply by building the response message (literally) */
	protected Document respondLiterally( File responseMessage, XMLMessage input) throws IOException, ParsingException
	{
		return xomBuilder.build( responseMessage );
	}
	
	
	/** Returns the base directory where pre-written response messages are stored */
	public static File getInstanceFileDir() throws IOException
	{
		return FrameworkSetup.getUserDirectory("xmlvend/test-server/responses");
	}
	
	/** Sets up the base directory - if the directory contains no content, it unpacks
	 * the default set of pre-written response files.
	 */
	public static void setupInstanceFileDir() throws IOException
	{
		if (getInstanceFileDir().list().length == 0)
		{
			logger.debug("No instance files in responses directory, unpacking default set");
			// Unzip default set of schemas
			ZipInflator.inflateZipStream( 
				BaseRequestProcessor.class.getClassLoader().getResourceAsStream("xmlvend-responses-default.zip"),
				getInstanceFileDir() );
		}
	}
	
	private static Logger logger = Logger.getLogger(BaseRequestProcessor.class);
	private SchemaTypeRegistry typeRegistry;
	private MessageTypeExtractor typeExtractor;
	private ResponseMap responses;
	private Builder xomBuilder = new Builder();
	private LastResponseStore lastResponseStore;
	private LastResponseAssembler lastResponseAssembler = new LastResponseAssembler();
	private CommonNameClientIDCheck cnCheck = new CommonNameClientIDCheck();
}
