package za.co.eskom.nrs.xmlvend.server;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Node;
import nu.xom.ParsingException;
import nu.xom.ProcessingInstruction;
import org.apache.log4j.Logger;
import za.co.solms.framework.ooxml.QualifiedName;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import za.co.solms.framework.ooxml.types.TypeNotFoundException;

/** Maps request types to pre-written response messages. Reads all the XML files 
 * recursively in a directory, and matches them to 'request type' by means of a 
 * processing instruction within the XML.
 * 
 * TODO This class needs work in terms of robustness during run-time scan() operations
 * 
 */
public class ResponseMap
{
	/** Builds a response map from all the files in the given  directory */
	public ResponseMap(File baseDirectory, SchemaTypeRegistry typeSystem) throws IOException
	{
		logger.info("Setting up pre-written response messages");
		this.registry = typeSystem;
		this.responseDir = baseDirectory;
		// Find responses and add
		scan();
		logger.info("Service is ready to respond with " + responseMap.size() + " message(s)" );
	}
	
	/** Scans for response messages */
	public void scan() throws IOException
	{
		// Lock the response map
		synchronized (responseMap)
		{
			addFilesInDir( responseDir );
			
			// TODO Remove any stale links (non-existent files)
			//...
		}
	}

	
	/** Finds the response (file) for the given request type */
	public File getResponse( SchemaType type )
	{
		logger.debug("Locating response for type: " + type);
		File f = responseMap.get( type );
		
		SchemaType t = type; // Type we're currently looking at
		// Walk up the type hierarchy
		while (f == null && t.getSuperType() != null)
		{
			t = t.getSuperType();
			logger.debug("No direct response for type, abstracting to: " + t);
			f = responseMap.get( t );
		}

		logger.debug("Response found: " + f);
		return f;
	}
	
	
	/** Recursively locates all XML files, and calls for them to be added */
	protected void addFilesInDir( File dir ) throws IOException
	{
		logger.debug("Reading files in directory: " + dir);
		File[] children = dir.listFiles( new XMLFilter() );
		for (File f : children)
		{
			if (f.isDirectory())
			{
				addFilesInDir( f );
			}
			else
			{
				addFile( f );
			}
		}
	}

	/** Adds the given XML files - reads the file, and finds the processing instruction */
	protected void addFile( File xmlFile) throws IOException
	{
		logger.debug("Adding response: " + xmlFile);
		// Read file (XOM)
		try
		{
			Document doc = xomBuilder.build( xmlFile );
		
			// Find Processing Instructions
			for (int x = 0; x < doc.getChildCount(); x++)
			{
				Node n = doc.getChild( x );
				
				// If it's our soap-response instruction, get the details (type)
				if (n instanceof ProcessingInstruction)
				{
					ProcessingInstruction pi = (ProcessingInstruction)n;
					String target = pi.getTarget();
					logger.debug("Found PI: " + target);
					
					if (target.equals( PROCESSING_INSTRUCTION_TARGET ))
					{
						try
						{
							logger.debug("Value: " + pi.getValue() );
							logger.debug("Resolving request type");
							// TODO Implement
							
							// Type Name
							String requestTypeN = null;
							Matcher typeNameMatcher = PI_TYPE_NAME_PATTERN.matcher( pi.getValue() );
							if (typeNameMatcher.find())
							{
								requestTypeN = typeNameMatcher.group(1);
							}
							
							// Type Namespace
							String namespaceN = null;
							Matcher typeNamespaceMatcher = PI_NAMESPACE_PATTERN.matcher( pi.getValue() );
							if (typeNamespaceMatcher.find())
							{
								namespaceN = typeNamespaceMatcher.group(1);
							}
							
							// If we can't derive a namespace, use the namespace
							// of the root element
							// FIXME: This is wrong, rather use document default namespace!!
							if (namespaceN == null)
							{
								namespaceN = doc.getRootElement().getNamespaceURI();
							}
							
							// Create type
							QualifiedName requestTypeName = new QualifiedName(namespaceN, requestTypeN );
							logger.debug("Request type: " + requestTypeName);
							
							try
							{
								SchemaType requestType = registry.getTypeByName( requestTypeName );
								
								// Add to map
								responseMap.put( requestType, xmlFile );
								
							} catch (TypeNotFoundException e)
							{
								logger.error("Response message (" + xmlFile + ") references unknown type: " + requestTypeName + " (ignoring)");
							}
						}
						catch (Exception e)
						{
							logger.error( xmlFile + " contains invalid processing instruction", e);
						}
					}
				}
			}
		}
		catch (ParsingException e)
		{
			logger.error("Invalid response file (XML Parsing failed)", e);
		}
	}
	
	private SchemaTypeRegistry registry;
	Logger logger = Logger.getLogger( ResponseMap.class );
	private Map<SchemaType,File> responseMap = new ConcurrentHashMap<SchemaType,File>();
	private Builder xomBuilder = new Builder();
	public static final String PROCESSING_INSTRUCTION_TARGET = "soap-response-to";
	public static final Pattern PI_TYPE_NAME_PATTERN = Pattern.compile("requestType[\\s]*=[\\s]*['\"]([^'\"]+)['\"]");
	public static final Pattern PI_NAMESPACE_PATTERN = Pattern.compile("namespace[\\s]*=[\\s]*['\"]([^'\"]+)['\"]");
	private File responseDir;
	
	/** Filter for xml files and directories */
	public static class XMLFilter implements FileFilter
	{
		public boolean accept(File file)
		{
			// Support common XML and XSLT extensions
			return file.isDirectory() || file.getName().toLowerCase().endsWith(".xml")
			|| file.getName().toLowerCase().endsWith(".xsl")
			|| file.getName().toLowerCase().endsWith(".xslt");
		}
	}
}
