package za.co.eskom.nrs.xmlvend.server.faults;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import org.apache.log4j.Logger;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.ParsingException;
import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessageFactory;
import za.co.solms.framework.ooxml.soap.processing.SOAPFaultFactory;

/** Produces faults for the XMLVend reference server, based on specific exceptions.
 * Functions similar to @link BaseRequestProcessor, in that response templates are
 * used as XSLT style sheets, with the original request as source document.
 */
public class XMlVendFaultFactory implements SOAPFaultFactory
{
	public XMlVendFaultFactory( File baseDir ) throws IOException
	{
		this.faultMap = new XSLTFileFaultMap( baseDir );
	}

	public SOAPMessage createFaultMessage(Throwable exception,
			SOAPMessage request)
	{
		logger.debug("Creating fault message for: " + exception.getClass().getName());
		// Get template for exception type
		File template = faultMap.getFaultTemplate( exception.getClass() );
		
		// Transform request using this template as XSLT
		Document response = null;
		try
		{
			logger.debug("Creating fault with template: " + template);
			
			if (template == null)
			{
				throw new NullPointerException("No fault template");
			}
			response = respondByTransformation( template, request, exception );
		}
		catch (Exception e)
		{
			logger.error("Fault response creation failed", e);
			// TODO: Rather throw exception
			return null;
		}
		
		// Return response (wrap in SOAP message)
		return SOAPMessageFactory.getInstance().createMessage( response.getRootElement() );
	}
	
	/** Create fault response message by transforming the input with the response. */
	protected Document respondByTransformation( File responseTemplate, XMLMessage input, Throwable exception) 
	throws TransformerException, IOException, ParsingException
	{
		// Setup XSLT
		logger.debug("Creating XSL transformer");
		// TODO Optimise this (cache)
		Transformer transformer = TransformerFactory.newInstance().newTransformer( new StreamSource(responseTemplate) );

		
		// Setup input as a stream (this is very inefficient)
		byte[] xmlData = input.getMessage().toXML().getBytes();
		ByteArrayInputStream bis = new ByteArrayInputStream(xmlData);
		
		// Setup output as a stream
		ByteArrayOutputStream bos = new ByteArrayOutputStream(256);

		// Make response date/time available
		Date now = new Date();
        SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'hh:mm:ss");
        String respDateTime = formatter.format(now);
        transformer.setParameter("Date", respDateTime );
        
        // Make details of exception available
        if (exception != null)
        {
        		transformer.setParameter( "exception.class", exception.getClass().getName() );
        		transformer.setParameter( "exception.message", exception.getMessage() );
        		// ...?
        }
		
		// Transform input document using response
		logger.debug("Transforming request using response");
		transformer.transform( new StreamSource(bis), new StreamResult(bos));
		byte[] data = bos.toByteArray();
		
		logger.debug("Done (ready for sending to client - " + data.length + " bytes)");

		// Create XOM tree for result (also very inefficient!)
		return xomBuilder.build( new ByteArrayInputStream( data ));
	}
	
	// Lcoates out fault templates
	private FaultMap faultMap;
	private static Logger logger = Logger.getLogger( XMlVendFaultFactory.class );
	private Builder xomBuilder = new Builder();
	
}