package za.co.eskom.nrs.xmlvend.server.faults;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Node;
import nu.xom.ParsingException;
import nu.xom.ProcessingInstruction;
import org.apache.log4j.Logger;
import za.co.eskom.nrs.xmlvend.server.ResponseMap.XMLFilter;

/** Recursively discovers XSLT templates with annotations (processing instructions)
 * that indicate the exception they are to be sent as a response to.
 * 
 * @author dawidl
 * @version 0.1 (Jun 13, 2006)
 */
public class XSLTFileFaultMap implements FaultMap
{
	/** Create by scanning for templates (recursively) in the
	 * given base directory.
	 */
	public XSLTFileFaultMap( File baseDir ) throws IOException
	{
		this.baseDir = baseDir;
		scan();
	}

	public void scan() throws IOException
	{
		logger.debug("Finding fault responses in directory: " + baseDir);
		synchronized (map)
		{
			addFilesInDir( baseDir );
			
			// TODO Remove any stale links (non-existent files)
			//...
		}
	}
	
	/** Recursively locates all XML files, and calls for them to be added */
	protected void addFilesInDir( File dir ) throws IOException
	{
		logger.debug("Reading files in directory: " + dir);
		File[] children = dir.listFiles( new XMLFilter() );
		for (File f : children)
		{
			if (f.isDirectory())
			{
				addFilesInDir( f );
			}
			else
			{
				addFile( f );
			}
		}
	}
	
	
	/** Adds the given XML files - reads the file, and finds the processing instruction */
	protected void addFile( File xmlFile) throws IOException
	{
		logger.debug("Examining fault response: " + xmlFile);
		// Read file (XOM)
		try
		{
			Document doc = xomBuilder.build( xmlFile );
		
			// Find Processing Instructions
			for (int x = 0; x < doc.getChildCount(); x++)
			{
				Node n = doc.getChild( x );
				
				// If it's our soap-response instruction, get the details (type)
				if (n instanceof ProcessingInstruction)
				{
					ProcessingInstruction pi = (ProcessingInstruction)n;
					String target = pi.getTarget();
					logger.debug("Found PI: " + target);
					
					if (target.equals( PROCESSING_INSTRUCTION_TARGET ))
					{
						try
						{
							logger.debug("Value: " + pi.getValue() );
							logger.debug("Validating exception type");
							
							// Type Name
							String exceptionTypeN = null;
							Matcher typeNameMatcher = PI_TYPE_NAME_PATTERN.matcher( pi.getValue() );
							if (typeNameMatcher.find())
							{
								exceptionTypeN = typeNameMatcher.group(1);
							}
							
							// Examine type
							try
							{
								// Validate class
								Class exceptionClass = Class.forName( exceptionTypeN );
								
								// TODO: Validate that this class is in fact throwable
								//...
								
								// Map file to it
								map.put( exceptionClass, xmlFile );
								logger.debug("Mapped " + exceptionTypeN + " to fault template: " + xmlFile);
							}
							catch (ClassNotFoundException e)
							{
								logger.warn("SOAP Fault template '" + xmlFile + "' references unkown exception type '" + exceptionTypeN + "' - ignoring");
							}
						}
						catch (Exception e)
						{
							logger.error( xmlFile + " contains invalid processing instruction", e);
						}
					}
				}
			}
		}
		catch (ParsingException e)
		{
			logger.error("Invalid response file (XML Parsing failed)", e);
		}
	}
	

	public File getFaultTemplate(Class exceptionClass)
	{
		/** Find map entry for given class (literally) or at any level fo abstraction */
		Class c = exceptionClass;
		File template = map.get( c );
		
		while (template == null && c.getSuperclass() != null)
		{
			c = c.getSuperclass();
			template = map.get( c );
		}
		
		return template;
	}

	// Stores class-to-file mappings
	private Map<Class,File> map = new HashMap<Class,File>();
	private File baseDir;
	private static Logger logger = Logger.getLogger( XSLTFileFaultMap.class );
	private Builder xomBuilder = new Builder();
	public static final String PROCESSING_INSTRUCTION_TARGET = "soap-fault-for";
	public static final Pattern PI_TYPE_NAME_PATTERN = Pattern.compile("exceptionType[\\s]*=[\\s]*['\"]([^'\"]+)['\"]");
}
