package za.co.solms.framework.impl.ooxml.basic.processing;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import org.apache.log4j.Logger;
import za.co.solms.framework.impl.ooxml.basic.types.MessageTypeExtractor;
import za.co.solms.framework.ooxml.QualifiedName;
import za.co.solms.framework.ooxml.XMLElementReference;
import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.processing.InvalidXMLProcessorException;
import za.co.solms.framework.ooxml.processing.UnhandledMessageTypeException;
import za.co.solms.framework.ooxml.processing.XMLInstance;
import za.co.solms.framework.ooxml.processing.XMLProcessingEngine;
import za.co.solms.framework.ooxml.processing.XMLProcessor;
import za.co.solms.framework.ooxml.processing.XMLType;
import za.co.solms.framework.ooxml.types.SchemaLocator;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;

/** This is a basic implementation of an XML Processing engine (which delegates XML
 * request messages to internally registered processors, as annotated to respond to
 * particular request types or instance names.
 * 
 * TODO IMPORTANT! This implementation only supports message routing by TYPE,
 * and not by INSTANCE NAME (even though such handlers are correctly registered).
 * This is because the current needs of this implementation do not require such
 * functionality, which is indeed unique to XML in anyway (polymorphic OO method 
 * arguments are typically dispatched by TYPE only, 'name' is unique to XML)
 * 
 * @author dawidl
 * @version 0.1 (Nov 10, 2005)
 *
 */
public class BasicXMLProcessingEngine implements XMLProcessingEngine
{
	/** Create a processing engine, which reads schemas from the given schema 
	 * locator to build a type registry. This constructor will build the
	 * type registry in a blocking fashion, so it may take a long time to
	 * complete.
	 */
	public BasicXMLProcessingEngine( SchemaLocator schemaLocator ) throws Exception
	{
		// Create a schema type registry
		logger.debug("Setting up schema catalog");
		typeRegistry = SchemaTypeRegistry.Factory.getInstance();
		
		// Add all the schema documents to it
		// This relies on the registry's built-in mechanisms to not load
		// duplicates (which there is bound to be, because some schemas import
		// each other) multiple times
		for (URL u : schemaLocator.getSchemaURLs())
		{
			typeRegistry.addTypesFromSchema( u );
		}
		
		// Calll init on ourself
		init( typeRegistry );
	}
	
	
	public void init(SchemaTypeRegistry typeRegistry)
	{
		// Type builder utility
		typeBuilder = new MessageTypeExtractor( typeRegistry );		
	}

	
	public void addProcessor(XMLProcessor processor) throws InvalidXMLProcessorException
	{
		logger.debug("Registering handler: " + processor);
		
		// If processor id invalid, don't bother
		if (processor == null)
			return;
		
		// Get all annotated methods
		for (Method m : processor.getClass().getMethods())
		{
			logger.debug("Looking at method: "+ m.getClass().getName() + 
					"." + m.getName());
			try
			{
				// Check if method signature is correct
				// Return Type
				if ( XMLMessage.class.equals( m.getReturnType() ))
				{
					// Parameters
					Class[] paramTypes = m.getParameterTypes();
					if (paramTypes.length == 1 && XMLMessage.class.equals (paramTypes[0]))
					{
						logger.debug("Method has correct signature");
						// Method signture is correct, now look at annotations
						for (Annotation a : m.getAnnotations())
						{
							logger.debug("Found annotation: " + a.getClass().getName());
							
							// Does this annotation handle XML Type?
							if (a instanceof XMLType)
							{
								// Construct type
								XMLType t = (XMLType)a;
								QualifiedName tq = new QualifiedName(t.nameSpace(), t.name());
								
								logger.debug("Handler for type: " + t + " : " + processor);
								
								try
								{
									// Check if we have this type
									SchemaType type = typeRegistry.getTypeByName( tq );
									
									// If so, register the new processing method
									ProcessorProxy pp = new ProcessorProxy( processor, m );
									// First initialise it (but only once ever)
									if (! initedHandlers.contains( processor.getClass() ))
									{
										logger.debug("Initialising processor");
										pp.init( typeRegistry );
										initedHandlers.add( processor.getClass() );
									}
									// Then add it
									typeHandlers.put( type, pp);
								} catch (Exception e)
								{
									throw new InvalidXMLProcessorException( processor, e);
								}
							}
							
							// Does this annotation handle XML Instance?
							if (a instanceof XMLInstance)
							{
								// Construct instance (root element name)
								XMLInstance i = (XMLInstance)a;
								QualifiedName iq = new QualifiedName( i.nameSpace(), i.name());
								
								logger.debug("Handler for instance: " + iq + " : " + processor);
								
								// Check if we have this as a valid root document
								try
								{
									XMLElementReference element = typeRegistry.getRootElementByName(iq);
									// (This is only to check, we still bind using name)
									
									// If so, register processing method
									instanceHandlers.put( iq, new ProcessorProxy( processor, m) );
								} catch (Exception e)
								{
									throw new InvalidXMLProcessorException( processor, e);
								}
							}
						}
					}
				}
				
			}
			catch (Exception e)
			{
				logger.error("Failed to register handler method", e);
			}
		}
	}

	
	public void removeProcessor(XMLProcessor processor)
	{
		logger.debug("Requested to remove processor: " + processor);
		if (processor == null) return;
		
		// Find the entry sets in the map(s), and remove if necessary
		for (Map.Entry<SchemaType,ProcessorProxy> entry : typeHandlers.entrySet())
		{
			if (entry.getValue().getProcessor() == processor || entry.getValue().getProcessor().equals( processor ))
			{
				typeHandlers.remove( entry.getKey() );
				logger.debug("Removed processor: " + processor);
			}
		}
		for (Map.Entry<QualifiedName,ProcessorProxy> entry : instanceHandlers.entrySet())
		{
			if (entry.getValue().getProcessor() == processor || entry.getValue().getProcessor().equals( processor ))
			{
				instanceHandlers.remove( entry.getKey() );
				logger.debug("Removed processor: " + processor);
			}
		}
		
		initedHandlers.remove( processor.getClass() );
	}

	
	/**
	 * Accepts a message for processing by an underlying processor. NOTE: Currently
	 * only processes by TYPE, not by NAME.
	 * 
	 *  @throws UnhandledMessageTypeException If the given message has no suitable processor
	 *  @throws za.co.solms.framework.ooxml.types.TypeNotFoundException If message type is unknown
	 * @see za.co.solms.framework.ooxml.processing.XMLProcessor#process(za.co.solms.framework.ooxml.XMLMessage)
	 */
	public XMLMessage process(XMLMessage input) throws Exception
	{
		logger.debug("Message received");
		// BY ELEMENT
		// Get FQ Name of element
		//... TODO In a later implementation, this implementation only supports BY TYPE
		
		// BY TYPE
		// Get Type of element
		SchemaType elementType = typeBuilder.deriveType( input.getMessage().getRootElement() );
		logger.debug("Message type: " + elementType);
		
		// Set type on element
		input.setType( elementType );
		
		// Get the best matching proxy
		ProcessorProxy p = getProxyForType( elementType );
		
		// Invoke
		logger.debug("Invoking handler");
		return p.process( input );
	}
	
	
	/** Locates the correct proxy to invoke, based on the given type. Match is
	 * performed at the most specific level of abstraction possible (polymorphically).
	 * @param type
	 * @return
	 */
	private ProcessorProxy getProxyForType( SchemaType type ) throws UnhandledMessageTypeException
	{
		logger.debug("Locating processor (proxy) for type: " + type);
		ProcessorProxy p = typeHandlers.get( type );
		SchemaType t = type; // Type we're currently looking at
		
		// Walk up the type hierarchy if needs be
		while (p == null && t.getSuperType() != null)
		{
			t = t.getSuperType();
			logger.debug("No direct handler for type, abstracting to: " + t);
			p = typeHandlers.get( t );
		}

		// Did we find a processor?
		if (p == null)
		{
			String msg = "No handler for type: " + type;
			logger.warn(msg);
			throw new UnhandledMessageTypeException(msg);
		}
		
		logger.debug("Handler found: " + p);
		return p;
	}

	
	public SchemaTypeRegistry getTypeSystem()
	{
		return typeRegistry;
	}
	
	// Our schema type registry
	SchemaTypeRegistry typeRegistry;
	// Mapping of methods by type
	Map<SchemaType, ProcessorProxy> typeHandlers = new  ConcurrentHashMap<SchemaType, ProcessorProxy>();
	// Mapping of methods by instance name
	Map<QualifiedName, ProcessorProxy> instanceHandlers = new ConcurrentHashMap<QualifiedName, ProcessorProxy>();
	// Utility to extract message types
	MessageTypeExtractor typeBuilder;
	// Log4J Logging
	private static Logger logger = Logger.getLogger(BasicXMLProcessingEngine.class);
	// Set of handler classes to keep track of which ones we've already initialised
	private Set<Class> initedHandlers = new HashSet<Class>();

	
	/** A proxy to a processing method in a processor */
	private class ProcessorProxy implements XMLProcessor
	{
		/** Creates a proxy, which enables the later invocation fo the processing method.
		 * @param processor
		 * @param method
		 */
		public ProcessorProxy(XMLProcessor processor, Method method)
		{
			this.processor = processor;
			this.method = method;
		}
		
		
		/** Initialises the underlying processor */
		public void init(SchemaTypeRegistry typeRegistry) throws Exception
		{
			processor.init( typeRegistry );
		}
		
		/** Invokes the XML Processing method */
		public XMLMessage process( XMLMessage input ) throws Exception
		{
			Object[] params = { input };
			try
			{
				return (XMLMessage)method.invoke( processor, params);
			}
			catch (InvocationTargetException e)
			{
				/** If invocation with reflection caused an exception, first unwrap
				 * it before re-throwing it.
				 */
				if (e.getCause() != null && e.getCause() instanceof Exception)
				{
					throw (Exception)e.getCause();
				}
				else
				{
					throw e;
				}
			}
		}
		
		/** Gets the underlying processor */
		protected XMLProcessor getProcessor()
		{
			return processor;
		}
		
		private XMLProcessor processor;
		private Method method;
	}
}
