package za.co.solms.framework.impl.ooxml.basic.types;

import java.io.IOException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import za.co.solms.framework.ooxml.*;
import za.co.solms.framework.ooxml.types.ElementNotFoundException;
import za.co.solms.framework.ooxml.types.SchemaConflictException;
import za.co.solms.framework.ooxml.types.SchemaException;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import za.co.solms.framework.ooxml.types.TypeInUseException;
import za.co.solms.framework.ooxml.types.TypeNotFoundException;

/** A basic implementation of a schema type registry. It uses the classloader to find the
 * W3C Schema for Schemas (in the package za.co.solms.framework.impl.ooxml.basic.schemas)
 * and load it at construction time. This is cached (statically) so that only the initial 
 * construction (first instance) is an expensive operation.
 */

public class BasicSchemaTypeRegistry implements SchemaTypeRegistry
{
	/** Constructs a registry. Loads all W3C schema types. */
	public BasicSchemaTypeRegistry() throws Exception
	{
		if (builtInTypes == null)
		{
			try
			{
				// Create store of built-in types
				builtInTypes = new HashSet<SchemaType>(100);
				// Add the "magic" anySimpleType base primitive type
				addType( new SchemaType( new QualifiedName( W3C_SCHEMA_NAMESPACE, "anySimpleType")) );
				// Populate with built-in types from W3C Schema
				URL schema4schema = getClass().getClassLoader().getResource("za/co/solms/framework/impl/ooxml/basic/schemas/XMLSchema.xsd");
				loader.loadTypesFromSchema( schema4schema, this);
			} catch (Exception e)
			{
				// If something went wrong, ensure that it won't prevent next 
				// time from trying (i.e. construction of new type registry)
				builtInTypes = null;
				throw e;
			}
		}
	}
	

	public SchemaType getTypeByName( QualifiedName typeName )
			throws TypeNotFoundException
	{
		// Not sure if this is bad for performance, but we create a type, and use
		// our helper method to look it up.
		return this.getType( new SchemaType( typeName ) );
	}

	
	public void addType(SchemaType type)
	{
		// If the type is in the schema namespace, add to built-in, otherwise add to user types
		if ( isNativeSchema( type.getName() ))
		{
			builtInTypes.add(type);
		}
		else
		{
			types.add(type);
		}
	}

	
	public void addTypesFromSchema(URL url) throws IOException, SchemaException
	{
		// Delegate to the type loader
		loader.loadTypesFromSchema( url, this);
	}

	
	public void removeType(SchemaType type) throws TypeInUseException, TypeNotFoundException
	{
		// If it's a built-in type, don't remove it
		if (isNativeSchema( type.getName() ) )
		{
			throw new TypeInUseException("Removal of built-in types not allowed");
		}
		
		// Get reference to our internal type
		SchemaType t = getType( type );
		
		// See if anything depends on this type
		// TODO Improve this very basic and non-failsafe method
		if (t.getKnownSubTypes().size() > 0)
		{
			throw new TypeInUseException("Type cannot be removed, it's extended by: " + t.getKnownSubTypes());
		}
		
		// Remove from registry
		types.remove( t );
		
		// If it has a supertype, unlink it
		if (t.getSuperType() != null)
		{
			t.getSuperType().removeSubType( t );
		}
	}
	
	
	public SchemaType getTypeFromRootElement(QualifiedName elementName) throws TypeNotFoundException
	{
		// Find matching root element, and get type
		for (XMLElementReference e : rootElements)
		{
			if (e.getName().equals( elementName ))
			{
				return e.getType();
			}
		}
		// Otherwise throw exception
		throw new TypeNotFoundException( elementName.toString() );
	}

	
	public XMLElementReference getRootElementByName(QualifiedName elementName) throws ElementNotFoundException
	{
		// Find matching root element, and get type
		for (XMLElementReference e : rootElements)
		{
			if (e.getName().equals( elementName ))
			{
				return e;
			}
		}
		// Otherwise throw exception
		throw new ElementNotFoundException( elementName.toString() );
	}

	/** Assigns a root element to a specific type. If the root element itself is in the Schema namespace,
	 * it's not added (ignored). 
	 * @see za.co.solms.framework.ooxml.types.SchemaTypeRegistry#assignRootElementToType(za.co.solms.framework.ooxml.QualifiedName, za.co.solms.framework.ooxml.types.SchemaType)
	 */
	public void assignRootElementToType(QualifiedName rootElementName, SchemaType type) throws TypeNotFoundException, SchemaConflictException
	{
		if (isNativeSchema( rootElementName ))
		{
			// TODO Is it a good idea to ignore root (built-in) types?
			return;
		}
		
		// See if we already have a root element of this name (if so, throw error, as that is ambiguous)
		for (XMLElementReference e : rootElements)
		{
			if (e.getName().equals( rootElementName ))
			{
				throw new SchemaConflictException("Element " + rootElementName + " is already assigned to another type: " + e.getType().getName());
			}
		}
		
		// Get handle to actual type
		SchemaType t = getType( type );
		
		// Create the root element and add to set of root elements
		XMLElementReference newElement = new XMLElementReference( rootElementName, t);
		rootElements.add( newElement );
	}
	
	
	/** Returns the type instance from the registry which is equal to the given type, otherwise
	 * returns null */
	protected SchemaType getType( SchemaType type ) throws TypeNotFoundException
	{
		for (SchemaType t :  (isNativeSchema( type.getName() ) ? builtInTypes : types) )
		{
			if (t.equals( type ))
			{
				return t;
			}
		}
		// If not found, throw exception
		throw new TypeNotFoundException("Type not in registry: " + type);
	}
	
	@Override
	public String toString()
	{
		String ts = "BasicSchemaTypeRegistry (STCD) ["+ types.size() +"T/"+rootElements.size()+"r]";
		return ts;
	}
	
	/** Check to see whether a type or element is built-in (XML Schema Native), or not. */
	protected boolean isNativeSchema( QualifiedName name )
	{
		return name != null && name.getNameSpace() != null && name.getNameSpace().equals(W3C_SCHEMA_NAMESPACE);
	}
	
	
	// Our store of user types. Concurrency-safe, at the cost of slower modification
	private Set<SchemaType> types = new CopyOnWriteArraySet<SchemaType>();
	
	// Our static store of built-in types (XML Schema Namespace). Meant to be populated only once, and in a controlled manner,
	// which means we can use a simpler, faster collection type
	private static Set<SchemaType> builtInTypes = null;
	
	// Our store of root elements. Concurrency-safe, at the cost of slower modification
	private Set<XMLElementReference> rootElements = new CopyOnWriteArraySet<XMLElementReference>();
	
	// Our loader, responsible for parsing schemas
	private SchemaTypeLoader loader = new SchemaTypeLoader();
	
	/** Schema namespace */
	public static final String W3C_SCHEMA_NAMESPACE = "http://www.w3.org/2001/XMLSchema";
}
