package za.co.solms.framework.impl.ooxml.basic.types;

import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import org.apache.log4j.Logger;
import za.co.solms.framework.ooxml.types.SchemaLocator;

/** A locator of schemas on the file system. Finds all .xsd files recursively
 * below a given starting directory. Contents are not cached, and the directory
 * will be searched for schemas upon every call to 'getSchemaURLs()'.
 * 
 * @author dawidl
 * @version 0.1 (Nov 29, 2005)
 */
public class FileSystemSchemaLocator implements SchemaLocator
{
	/** Create a schema locator for all schemas below the given directory.
	 * @throws FileNotFoundException If the directory does not exist
	 */
	public FileSystemSchemaLocator( File baseDirectory ) throws FileNotFoundException
	{
		if (baseDirectory == null || !baseDirectory.exists() 
				|| !baseDirectory.isDirectory())
		{
			throw new FileNotFoundException("Schema Directory does not exist: " 
					+ baseDirectory);
		}
		
		this.baseDirectory = baseDirectory;
		logger.debug("Created for base directory: " + baseDirectory);
	}

	
	/** Locates schemas recursively in the given directory.
	 * TODO Add short-term caching for performance? 
	 * @see za.co.solms.framework.ooxml.types.SchemaLocator#getSchemaURLs()
	 */
	public Set<URL> getSchemaURLs()
	{
		logger.debug("Schema URLs requested (locating...)");
		
		Set<URL> files = new HashSet<URL>();
		try
		{
			addFileURLsToList( files, baseDirectory );
		} catch (IOException e)
		{
			// TODO Interface does not allow for exception - what to do?
			logger.error("Unable to locate schemas", e);
		}
		
		return files;
	}
	
	
	/** Adds all files in the given directory to the list recursively */
	private void addFileURLsToList( Set<URL> files, File directory) throws IOException
	{
		logger.debug("Looking at directory: " + directory);
		File[] fListing = directory.listFiles( filter );
		for (File f : fListing)
		{
			if (f.isDirectory())
			{
				addFileURLsToList( files, f);
			}
			else
			{
				logger.debug("Found schema: " + f);
				files.add( f.toURL() );
			}
		}
	}

	private File baseDirectory;
	private SchemaFilter filter = new SchemaFilter();
	private Logger logger = Logger.getLogger( FileSystemSchemaLocator.class );
	
	
	/** Filter for schemas and directories */
	private class SchemaFilter implements FileFilter
	{
		public boolean accept(File file)
		{
			return file.isDirectory() || file.getName().toLowerCase().endsWith(".xsd");
		}
	}
}
