package za.co.solms.framework.impl.ooxml.basic.types;

import org.apache.log4j.Logger;

import nu.xom.Element;
import za.co.solms.framework.impl.ooxml.basic.util.NamespaceMap;
import za.co.solms.framework.impl.ooxml.basic.util.UnresolvableNamespaceException;
import za.co.solms.framework.ooxml.QualifiedName;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import za.co.solms.framework.ooxml.types.TypeNotFoundException;

/** A utility class to determine the type of an XML message (making use of
 * an XML Schema Type Registry) at any level of abstraction it knows of.
 * It allows for a more flexible implementation of a web services server,
 * which may receive messages of a specialised type, of which it only understands
 * a generic type (e.g. the client has extended their vocabulary with additional
 * schemas / types).
 * 
 * Types can only reliably be derived for ROOT ELEMENTS (as per the definition
 * in the SchemaTypeRegistry interface) as any other element is without context.
 * 
 * TODO: Abstract an interface for this functionality
 * 
 * @author dawidl
 * @version 0.1 (Nov 10, 2005)
 *
 */
public class MessageTypeExtractor
{
	/** Create a type extractor, using the given registry to look up
	 * root elements and schema types.
	 */
	public MessageTypeExtractor(SchemaTypeRegistry typeRegistry )
	{
		this.typeRegistry = typeRegistry;
	}
	
	
	/** Derives the TYPE of the given root element, based on the known types
	 * as provided by the registry. Returns a reference to the actual registry
	 * type. First examines xsi:type attribute, failing which, the type is looked
	 * up in the registry using the element name, as per valid root element
	 * definitions in the schema. NOTE: This service does in no way validate
	 * the type, i.e. it will blindly accept a valid xsi:type without checking for
	 * validity in respect to the element name - that is a separate validator's
	 * responsibility.
	 * 
	 * TODO This service needs significant additions, see comments in method body
	 * 
 	 * @param rootElement
	 * @return The schema type, as per the registry
	 * @throws TypeNotFoundException if the type cannot be derived.
	 */
	public SchemaType deriveType( Element rootElement ) throws TypeNotFoundException
	{
		// Basic checking
		if (rootElement == null)
		{
			throw new TypeNotFoundException( "null element" );
		}
		
		/* TODO Extensibility: As per discussions with Fritz et al, we need to add 
		 * code to automatically find any schemas for unknown namespaces (as per
		 * xsi:schemaLocation attribute) and load it, and then derive the type.
		 * This is to facilitate the web services vision of automatic inferrance,
		 * discovery, integration, etc...
		 */
			
		// Build a name space map (to resolve namespace prefixes for element or type)
		NamespaceMap nsMap = new NamespaceMap( rootElement );
		
		// By Type (Directly)
		// Get xsi:type, and expand
		String typeName = rootElement.getAttributeValue("type",XSI_NAMESPACE);
		if (typeName != null)
		{
			// Get the type
			try
			{
				QualifiedName typeQName = nsMap.expand( typeName, rootElement.getNamespaceURI() );
				return typeRegistry.getTypeByName( typeQName );
			} catch (UnresolvableNamespaceException e)
			{
				logger.error("Unable to derive type, invalid namespace prefix", e );
				throw new TypeNotFoundException( typeName );
			}
		}
		else
		{
			// By Name (Inferrance)
			String ns = rootElement.getNamespaceURI();
			String localName = rootElement.getLocalName();
			QualifiedName name = new QualifiedName(ns, localName);
			
			return typeRegistry.getTypeFromRootElement( name );
		}
	}

	private SchemaTypeRegistry typeRegistry;
	public static final String XSI_NAMESPACE = "http://www.w3.org/2001/XMLSchema-instance";
	private static Logger logger = Logger.getLogger( MessageTypeExtractor.class );
}
