package za.co.solms.framework.impl.ooxml.basic.types;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.apache.log4j.Logger;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.XPathContext;
import za.co.solms.framework.impl.ooxml.basic.util.NamespaceMap;
import za.co.solms.framework.ooxml.QualifiedName;
import za.co.solms.framework.ooxml.types.SchemaConflictException;
import za.co.solms.framework.ooxml.types.SchemaException;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;

/** This is a helper to parse a W3C schema at a known URL, extract all the types,
 * and populate a given SchemaTypeRegistry. This loader will traverse imported
 * schemas to an arbitrarily deep nested level, and protects against direct or indirect 
 * cyclic references.
 * 
 * A type loader does not maintain a reference to a schema type registry, so one can be used to
 * populate multiple registries if required.
 * 
 * TODO Cyclic dependency protection needs decent testing.
 * TODO Schema catalog / caching for networked loading?
 * 
 * @author dawidl
 * @version 0.1 (Oct 17, 2005)
 */

public class SchemaTypeLoader
{
	public SchemaTypeLoader(){}
	
	/** Loads the schema from the given URL, and populates the given registry.
	 * Will first load types from any further schemas 'imported' by the given schema. 
	 * If any of the schemas are currently being loaded as part of a recursive load call,
	 * it is simply ignored.*/
	public void loadTypesFromSchema( URL schemaURL, SchemaTypeRegistry registry) throws SchemaException, IOException
	{
		logger.debug("Loading schema from " + schemaURL + " into registry " + registry);
		
		// Check if url etc are valid
		if (schemaURL == null)
		{
			logger.error("Invalid schema URL (null)");
			throw new MalformedURLException("Invalid URL: null");
		}
		
		// If this schema is already being loaded as part of a recursive call, return immediately
		if (schemaURLLock.contains( schemaURL ))
		{
			logger.warn("Cyclic dependency or concurrent load: " + schemaURL + " (ignoring)");
			return;
		}
		
		// If this schema was previous loaded, do not reload
		if (loadedSchemas.contains( schemaURL ))
		{
			logger.debug("Schema has already been loaded: " + schemaURL + " (not reloading)");
			return;
		}
		
		logger.info("Loading schema: " + schemaURL);
		
		// Get Schema contents
		Document schemaDocument = getSchema( schemaURL );
		
		// Add load lock for this schema, to make sure imports don't indirectly cause it to be loaded again
		schemaURLLock.add( schemaURL );
		
		// First see if there are any imports, and add them (recursively, blocking)
		try
		{
			addImportsToRegistry( schemaDocument, registry);
		}
		catch (Exception e)
		{
			// TODO If one or schemas failed to import, should we continue?
			// In this case, we do continue
			logger.error("One or more referenced schemas failed to import", e);
		}
		
		try
		{
			// What is this schema's target name space (if any)?
			Nodes tns = schemaDocument.query("/*/@targetNamespace",xpContext);
			String targetNamespace = null;
			if (tns.size() > 0)
			{
				targetNamespace = tns.get(0).getValue();
				logger.debug("Target name space: " + targetNamespace);
			}
			else
			{
				logger.warn("Schema has no target name space: " + schemaURL);
			}
			
			// What are all the other namespaces used in this schema?
			NamespaceMap nsMap = new NamespaceMap( schemaDocument.getRootElement() );
			
			// The method of adding the types to the registry is a simple two-pass method
			// that first adds all types, then iterates through the schema types again
			// in order to link up the types (according to their OO relationships - 
			// extension/restriction etc. This is because types can appear out of order in
			// a schemas document, and we're cycling through all the types linearly. This is
			// not a particularly robust method (though it should be very fast) - the alternative
			// would be to find types on-demand in the schema, and effectively navigate the
			// types tree as represented in the schema document
			
			
			// Find all the types in this schema (add to list of types)
			List<Element> typeList = new ArrayList<Element>();
			Elements e = schemaDocument.getRootElement().getChildElements();
			for (int x = 0; x < e.size(); x++)
			{
				Element ee = e.get( x );
				if (ee.getLocalName().equals("complexType") || ee.getLocalName().equals("simpleType"))
				{
					typeList.add( ee );
				}
			}
			
			// FIRST PASS (add types to the registry)
			logger.debug("Adding types to registry");
			for (Element t : typeList)
			{
				String name = t.getAttributeValue("name");
				// Use our own schema type with settable supertype
				MutableSchemaType type = new MutableSchemaType( new QualifiedName(targetNamespace, name));
				// TODO Is this bad design?
				type.setTypeRegistry( registry );
				registry.addType( type );
				logger.debug("Added: " + type);
			}
			
			// SECOND PASS (relationships)
			logger.debug("Setting up type relationships");
			for (Element t : typeList)
			{
				String name = t.getAttributeValue("name");
				try
				{
					QualifiedName typeName = new QualifiedName(targetNamespace, name);
					logger.debug("Looking at type: " + typeName);
					
					// Find it in the registry
					SchemaType type = registry.getTypeByName( typeName );
					
					// Get Name of the 'extended' type
					String baseName = null;
					if (t.getLocalName().equals("complexType"))
					{
						// Look for complexContent
						Elements a = t.getChildElements("complexContent",SCHEMA_NS);
						if (a.size() > 0)
						{
							// Look for extensions or restriction
							Elements b = a.get(0).getChildElements();
							for (int x = 0; x < b.size(); x++)
							{
								Element c = b.get(x);
								if (c.getLocalName().equals("extension") || c.getLocalName().equals("extension"))
								{
									baseName = c.getAttributeValue("base");
								}
							}
						}
					}
					else if (t.getLocalName().equals("simpleType"))
					{
						// Look for extension or restriction
						Elements b = t.getChildElements();
						for (int x = 0; x < b.size(); x++)
						{
							Element c = b.get(x);
							if (c.getLocalName().equals("extension") || c.getLocalName().equals("restriction"))
							{
								baseName = c.getAttributeValue("base");
							}
						}
					}
					
					// If we have a base name, hook it up to the type in the registry
					if (baseName != null)
					{
						// Expand the name, and find the type in the registry
						// FIXME We blatantly and incorrectly assume here that if the
						// type is not prefixed, it must be a built-in schema type.
						QualifiedName baseQName = nsMap.expand( baseName, SCHEMA_NS);
						logger.debug(typeName + " is subtype of " + baseQName);
						SchemaType superType = registry.getTypeByName( baseQName );
						// Connect the two types (this relies on the type being of our mutable kind)
						((MutableSchemaType)type).setSuperType( superType );
					}
				} catch (Exception te)
				{
					logger.error("Unable to set up type: " + name, te);
				}
			}
			
			// ROOT ELEMENTS (Only ones with type references)
			Nodes roots = schemaDocument.getRootElement().query("/*/xs:element[@type]",xpContext);
			for (int x = 0; x < roots.size(); x++)
			{
				Element rootElement = (Element)roots.get(x);
				
				try
				{
					// Element name
					String name = rootElement.getAttributeValue("name");
					
					// Find the type name
					String typeRef = rootElement.getAttributeValue("type");
					
					// Expand, and get handle to type in registry
					QualifiedName expectedType = nsMap.expand( typeRef );
					SchemaType type = registry.getTypeByName( expectedType );
					
					// Add it to the registry
					QualifiedName rootElementName = new QualifiedName(targetNamespace, name);
					registry.assignRootElementToType(rootElementName , type );
					logger.debug("Root element: " + rootElementName + " (Type: " + type + ")");
					
				}
				catch (SchemaConflictException sce)
				{
					logger.warn( sce.getMessage() );
				}
				catch (Exception ge)
				{
					logger.error("Unable to add root element (skipping)", ge);
				}
			}
			
			
			
		}
		finally
		{
			// Remove load lock for this schema
			logger.debug("Removing load lock for: " + schemaURL.toExternalForm());
			schemaURLLock.remove( schemaURL );
		}
		
		// If all went well, add to list of successfully loaded schemas
		loadedSchemas.add( schemaURL );
	}
	
	
	/** Adds any imports found in the given schema (XOM) the registry. If no imports are found,
	 * it simly returns. It functions by resolving the URLs of the schemas, and calling the
	 * loadTypesfromSchema() method.
	 * @throws IOException If one or of the schemas could not be imported.
	 * @throws SchemaException If the given schema document is invalid
	 * @throws URISyntaxException If an invalid URI is contained within one of the documents 
	 * */
	protected void addImportsToRegistry(Document schemaDocument, SchemaTypeRegistry registry) throws SchemaException, IOException, URISyntaxException
	{
		if (schemaDocument == null)
		{
			throw new SchemaException("Invalid (null) schema: Unable to import references", null);
		}
		
		URL docBaseURL = new URI( schemaDocument.getBaseURI() ).toURL();
		
		logger.debug("Looking for imported schemas");
		Element rootElement = schemaDocument.getRootElement();
		// Get all imports (nodes) using XPath
		Nodes importNodes = rootElement.query("//xs:import", xpContext);
		// Cyle over them and see if they refer to other files (versus namespace-only imports)
		for (int c = 0; c < importNodes.size(); c++)
		{
			try
			{
				Node importNode = importNodes.get(c);
				
				// Get URL (otherwise it's a namespace-only import)
				Nodes location = importNode.query("@schemaLocation");
				if (location != null && location.size() > 0)
				{
					String uri = location.get(0).getValue();
					URL importURL = new URL( docBaseURL, uri);
					logger.debug("Found import: " + importURL);
					
					// Recursively import
					loadTypesFromSchema( importURL, registry );
				}
				else
				{
					// Namespace-only import - do we need to do anything here?
				}
			} catch (Exception e)
			{
				logger.error("Failed to import schema (skipping)", e);
			}
		}
	}
	
	
	/** Builds a XOM document from the Schema URL */
	protected Document getSchema(URL location) throws IOException, SchemaException
	{
		// TODO: Validation?? For now, it's false (we need the schema for schemas)
		boolean validation = false;
		
		// Read schema (using XOM)
		logger.debug("Parsing schema (validation=" + validation + ")");
		Builder parser = new Builder(validation);
		Document schemaDocument;
		try
		{
			String schemaURL = location.toExternalForm().replace(" ","%20");
			logger.debug("Schema URL: " + schemaURL);
			schemaDocument = parser.build(  schemaURL );
			return schemaDocument;
		}
		catch (IOException passthru)
		{
			throw passthru;
		}
		catch (Exception ve)
		{
			SchemaException se = new SchemaException("Schema parsing failed: " + location, ve);
			logger.error(se);
			throw se;
		}
	}
	
	
	/** Adds the node representing the <xs:complexType> to the given rgeistry */
	protected void addComplexTypeToRegistry(Node complexType, SchemaTypeRegistry registry)
	{
		
	}
	
	/** Adds the node representing the <xs:simpleType> to the given rgeistry */
	protected void addSimpleTypeToRegistry(Node simpleType, SchemaTypeRegistry registry)
	{
		
	}
	
	
	// Stores the URLs of schemas currently being loaded. Used to prevent cyclic loops
	private Set<URL> schemaURLLock = new HashSet<URL>();
	
	// Stores URL of schemas successfully parsed (in order to not re-load)
	private Set<URL> loadedSchemas = new HashSet<URL>();
	
	// Logging
	private static Logger logger = Logger.getLogger( SchemaTypeLoader.class );
	
	// XPath context for querying XML Schemas
	private static final String SCHEMA_NS = "http://www.w3.org/2001/XMLSchema";
	private static XPathContext xpContext = new XPathContext("xs",SCHEMA_NS);
	
	/** A schema type which allows the superType and registry to be set after creation time */
	private class MutableSchemaType extends SchemaType
	{
		public MutableSchemaType(QualifiedName typeName)
		{
			super( typeName );
		}
		
		/** Relax visibility constraint from superclass */
		@Override
		public void setSuperType( SchemaType type )
		{
			super.setSuperType( type );
		}
		
		public void setTypeRegistry( SchemaTypeRegistry typeRegistry)
		{
			this.typeRegistry = typeRegistry;
		}
	}
}