package za.co.solms.framework.impl.ooxml.basic.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import za.co.solms.framework.ooxml.QualifiedName;
import nu.xom.Node;

/** This class enables lookup operations for the name spaces in an XML document (either 
 * from String or XOM model). It does so without context (i.e. document global)
 * so it will behave unpredictably if different namespaces are bound to the same prefix (which
 * is generally a very bad idea to begin with, it makes documents very confusing).
 * 
 * This class also offers utility services to look up name spaces, etc.
 * 
 * NOTE: The current implementation of this class is based on String and Regular Expressions
 * for reasons of simplicity or speed. It does not enforce or check any XML well-formedness rules on
 * the document.
 * 
 * FIXME This class needs to support the document's "default" namespace
 * 
 * @author dawidl
 * @version 0.1 (Nov 1, 2005)
 */
public class NamespaceMap
{

	/** Creates a map for the given document. If the document contains no namespaces,
	 * this operation will not fail.
	 * @param xmlDocument
	 */
	public NamespaceMap(String xmlDocument)
	{
		if (xmlDocument != null && xmlDocument.length() > 0)
		{
			Matcher m = NS_MAPPING_PATTERN.matcher( xmlDocument );
			while (m.find())
			{
				String prefix = m.group(1);
				String namespace = m.group(2);
				nsMap.put(prefix, namespace);
			}
		}
	}
	
	
	/** Creates a map for the given XOM node */
	public NamespaceMap( Node xmlFragment )
	{
		// Call other constructor
		this ( xmlFragment != null ? xmlFragment.toXML() : null );
	}
	
	
	/** Returns the namespace mapped to the given prefix
	 * @throws UnresolvableNamespaceException If the namespace is unknown
	 * */
	public String getNamespaceForPrefix( String prefix )  throws UnresolvableNamespaceException
	{
		String ns = nsMap.get( prefix );
		if (ns == null)
		{
			throw new UnresolvableNamespaceException( prefix );
		}
		else
		{
			return ns;
		}
	}
	
	
	/** Returns the namespaceprefix(es) mapped to the given namespace. If no prefixes
	 * are found, an empty set is returned (never null) */
	public Set<String> getPrefixesForNamespace( String namespace )
	{
		Set<String> prefixes = new HashSet();
		// Cycle through all entries and add matches to the set
		for ( Entry<String,String> e : nsMap.entrySet() )
		{
			if (e.getValue().equals( namespace ))
			{
				prefixes.add( e.getKey() );
			}
		}
		return prefixes;
	}
	

	// @Component
	/** Returns a copy of all the namespaces. */
	public Map<String,String> getAllNamespaces()
	{
		return new HashMap<String, String>( nsMap );
	}
	
	
	/** A convenience method to return the expanded namespace of a prefixed element,
	 * e.g. "a:myElement". Also useful for schema types, etc.
	 * @param element The XML element (name, or optionally enclosed in brackets)
	 * @return
	 */
	public String expandPrefix( String prefixedName ) throws UnresolvableNamespaceException
	{
		if (prefixedName == null) return null;
		Matcher m = NS_PREFIX_PATTERN.matcher( prefixedName );
		if (m.find())
		{
			// Match first capture group
			return  getNamespaceForPrefix( m.group(1) );
		}
		else
		{
			return null;
		}
	}
	
	/** A convenience method to turn a prefixed name into a proper qualified name
	 * @param A name like "foo:Bar" 
	 * @return A QualifiedName, e.g. {http://foo/foo}Bar
	 * @throws UnresolvableNamespaceException If the prefix is unknown to this map
	 */
	public QualifiedName expand( String name ) throws UnresolvableNamespaceException
	{
		return expand(name, null);
	}
	
	
	/** A convenience method to turn a prefixed name into a proper qualified name,
	 * using the provided default namespace if the name is not prefixed.
	 * @param name A name like "foo:Bar"
	 * @param defaultNamespace if the name does not contain a prefix, use this namespace 
	 * @return A QualifiedName, e.g. {http://foo/foo}Bar
	 * @throws UnresolvableNamespaceException If the prefix is unknown to this map
	 */
	public QualifiedName expand( String name, String defaultNamespace ) throws UnresolvableNamespaceException
	{
		// Do we have a namespace?
		int c = name.indexOf(':');
		if ( c > 0)
		{
			// Expand prefix
			String nsPart = expandPrefix( name );
			// Get name part alone
			String namePart = name.substring( c + 1);
			return new QualifiedName(nsPart, namePart);
		}
		else
		{
			return new QualifiedName(defaultNamespace, name);
		}
	}

	
	// Stores name space mappings
	private Map<String, String> nsMap = new HashMap();
	
	// Document's default namespace
	private String defaultNamespace;
	
	/** Compiled Regex Patttern to match namespace declarations in XML, with two capture groups.
	 * First capture group captures prefix, second capture group captures namespace */
	public static final Pattern NS_MAPPING_PATTERN = Pattern.compile("xmlns:([\\w]+)[\\s]*=[\\s]*['\"]([^'\"]+)['\"]");
	
	/** Compiled regex pattern to match the namespace prefix of an element or type reference */
	public static final Pattern NS_PREFIX_PATTERN = Pattern.compile("([\\w]+):[\\w+]");
}
