package za.co.solms.framework.impl.ooxml.soap.processing;

import nu.xom.Element;

import org.apache.log4j.Logger;

import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.processing.InvalidXMLProcessorException;
import za.co.solms.framework.ooxml.processing.XMLProcessingEngine;
import za.co.solms.framework.ooxml.processing.XMLProcessor;
import za.co.solms.framework.ooxml.soap.SOAPMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessageFactory;
import za.co.solms.framework.ooxml.soap.processing.SOAPFaultFactory;
import za.co.solms.framework.ooxml.soap.processing.SOAPProcessor;
import za.co.solms.framework.ooxml.soap.processing.WSDLLocator;
import za.co.solms.framework.ooxml.types.SchemaLocator;
import za.co.solms.framework.ooxml.types.TypeNotFoundException;

/** A basic SOAP Processing Engine. It simply wraps around an OO XML Processing 
 * Engine, taking care of SOAP Wrapping/Unwrapping, as well as error semantics, etc.
 */
public class BasicSOAPProcessor implements SOAPProcessor
{

	public BasicSOAPProcessor(WSDLLocator wsdl, SchemaLocator schemas) throws Exception
	{
		logger.info("SOAP Processor Initialising (creating OOXML processor)");
		xmlProcessor = XMLProcessingEngine.Factory.getInstance( schemas );
		this.wsdlLocator = wsdl;
		
		// Fault factory (with default for backup if primary one fails)
		this.faultFactory = new DefaultSOAPFaultFactory();
		this.defaultFaultFactory = faultFactory;
		
		logger.info("SOAP Processor ready");
	}

	public SOAPMessage process(SOAPMessage request)
	{
		logger.debug("SOAP Message received");
		try
		{
			// Hand to processing engine (message unwraps itself)
			XMLMessage response = xmlProcessor.process( request );
			
			// If we don't have a response, send back null (asynchronous messaging)
			if (response == null)
			{
				return null;
			}
			// Check if we have a SOAP message (which we can send back directly)
			else if (response instanceof SOAPMessage)
			{
				logger.debug("Handler responded with SOAP message");
				return (SOAPMessage)response;
			}
			// Otherwise we need to wrap it in a SOAP envelope
			else
			{
				logger.debug("Handler responded with plain message, wrapping in SOAP");
				return SOAPMessageFactory.getInstance().createMessage( response );
			}
		}
		catch (Exception e)
		{
			logger.error("Message invocation failed",e);
			return createFaultMessage( e, request );
		}
	}
	
	
	private SOAPMessage createFaultMessage( Throwable exception, SOAPMessage request )
	{
		try
		{
			SOAPMessage fault = faultFactory.createFaultMessage( exception, request );
			if (fault == null)
			{
				throw new NullPointerException("Fault factory produced invalid (null) fault");
			}
			return fault;
		}
		catch (Exception e)
		{
			logger.error("Fault factory failed, using default", e);
			return defaultFaultFactory.createFaultMessage( exception, request );
		}
	}
	
	
	public void setFaultFactory(SOAPFaultFactory factory)
	{
		if (factory != null)
		{
			this.faultFactory = factory;
			logger.debug("Using SOAP Fault factory: " + factory);
		}
		else
		{
			this.faultFactory = new DefaultSOAPFaultFactory();
			logger.debug("Using default SOAP Fault factory");
		}
	}

	
	public void addProcessor(XMLProcessor processor) throws InvalidXMLProcessorException
	{
		xmlProcessor.addProcessor( processor );
	}
	
	public void removeProcessor(XMLProcessor processor) throws InvalidXMLProcessorException
	{
		xmlProcessor.removeProcessor( processor );
	}
	
	public WSDLLocator getWSDLLocator()
	{
		return wsdlLocator;
	}

	private XMLProcessingEngine xmlProcessor;
	private WSDLLocator wsdlLocator;
	private static Logger logger = Logger.getLogger( BasicSOAPProcessor.class );
	public static final String ACTOR_NAME = "urn:ooxml:soap-engine";
	private SOAPFaultFactory faultFactory, defaultFaultFactory;
	
	
	/** A basic implementation fo a exception-to-SOAP fault mapper */
	public static class DefaultSOAPFaultFactory implements SOAPFaultFactory
	{
		public SOAPMessage createFaultMessage(Throwable e, SOAPMessage request)
		{
			logger.debug("Creating SOAP Fault for exception: " + e.getClass().getName());
			
			// If this is a TypeNotFound exception, it's likely the client's fault
			String fCode, fString = null;
			if (e instanceof TypeNotFoundException)
			{
				fCode = "soap:Client";
				fString = "Server does not understand the XML type of client message: " + e.getMessage();
			}
			else
			{
				fCode = "soap:Server";
				fString = e.getMessage() != null ? e.getMessage() : "No Details (" + e.getClass().getName() + ")";
				
				// Add root cause
				if (e.getCause() != null)
				{
					fString += " caused by: " + e.getCause().toString();
				}
			}
			
			Element fault = new Element("Fault",SOAPMessageFactory.SOAP_NAMESPACE);
			Element faultcode = new Element("faultcode", "");
			faultcode.appendChild(fCode);
			fault.appendChild( faultcode );
			
			Element faultstring = new Element("faultstring", "");
			faultstring.appendChild(fString);
			fault.appendChild( faultstring );
			
			// TODO Fault actor and detail need to be clarified
			/*
			Element faultactor = new Element("faultactor", "");
			faultactor.appendChild(ACTOR_NAME);
			fault.appendChild( faultactor );
			Element detail = new Element("detail", "");
			*/
			
			return SOAPMessageFactory.getInstance().createMessage( fault );
		}
	}
}