package za.co.solms.framework.ooxml;

/** This class abstractly represents a Qualified Name - i.e. something which has both 
 * a name, and belongs to a name space. It can thus be used both for "Type" or "Class" names,
 * as well as for XML elements. This contrasts with the Java (JAX) QName, which is used typically
 * only for element names. Qualified names are immutable, i.e. cannot change once created.
 * TODO Is immutability a good idea?
 */
public class QualifiedName implements Comparable
{
	/** Create a qualified name. The name itself may never be null.
	 * @throws IllegalArgumentException if the name is null */
	public QualifiedName( String nameSpace, String name) throws IllegalArgumentException
	{
		if (name == null)
			throw new IllegalArgumentException("Attempt to create Qualified Name with null value for name");
		
		this.nameSpace = nameSpace;
		this.name = name;
	}
	
	
	public String getName()
	{
		return name;
	}
	
	
	public String getNameSpace()
	{
		return nameSpace;
	}

	
	/** Returns the qualified name
	 * @return A String in the format '{namespace}name', or just 'name' if there is no namespace.
	 */
	public String getQualifiedName()
	{
		String fullName = this.name;
		if (nameSpace != null)
		{
			fullName = "{" + nameSpace.toString() + "}" + fullName; 			
		}
		return fullName;
	}
	
	
	@Override
	public boolean equals(Object other)
	{
		// Instance check
		if (other == this)
		{
			return true;
		}
		
		// Attributes check
		try
		{
			QualifiedName o = (QualifiedName)other;
			return
				// Name must match (and will never be null)
				o.getName().equals( name ) 
				&&
				// Namespace must match. Either can be null
				(
					(nameSpace == null && o.getNameSpace() == null)
					||
					(nameSpace.equals( o.getNameSpace() ))
				);
		}
		catch (Exception e)
		{
			return false;
		}
	}
	
	
	@Override
	public String toString()
	{
		return this.getQualifiedName();
	}
	
	
	/** Compares to another abject, to determine natural ordering. Based on type name only */
	public int compareTo(Object other)
	{
		try
		{
			QualifiedName o = (QualifiedName)other;
			return o.getName().compareTo( name );
		}
		catch (Exception e)
		{
			return 0;
		}
	}
	
	
	@Override
	public int hashCode()
	{
		// Delegate to String
		return getQualifiedName().hashCode();
	}

	
	//@Component
	private String nameSpace, name;
}
