package za.co.solms.framework.ooxml.processing;

import za.co.solms.framework.impl.ooxml.basic.processing.BasicXMLProcessingEngine;
import za.co.solms.framework.ooxml.types.SchemaLocator;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;

/** A facade to one or more @link XMLProcessor instances. Several processors
 * can be plugged into a processing engine, and it will polymorphically dispatch
 * messages to the correct processing methods based on their annotations
 * (such as '@XMLInstance' and '@XMLType'). This is achieved by calling the
 * 'process()' service (which need not be annotated) of an implementation of
 * this interface. NOTE: This service will throw an UnhandledMessageTypeException
 * if no suitable processor is plugged in for a particular message.
 * 
 * Implementations of this interfaqce are responsible for managing their own
 * Object-Oriented XML sub-system (such as the implementation in the 
 * 'za.co.solms.framework.ooxml.types' package) in order to enable the polymorphic
 * behaviour.
 * 
 * 
 * @author dawidl
 * @version 0.1 (Nov 6, 2005)
 *
 */
public interface XMLProcessingEngine extends XMLProcessor
{

	/** Adds a processor, which is immediately available for polymorphic
	 * XML mesage dispatch.
	 * TODO Add exceptions to the contract
	 * @param processor
	 */
	public void addProcessor( XMLProcessor processor ) throws InvalidXMLProcessorException;
	
	
	/** Removes a processor. Any operations in progress, or operations
	 * queued internally (already dispatched to this processor) must be allowed 
	 * to complete normally.
	 * @param processor
	 */
	public void removeProcessor( XMLProcessor processor ) throws InvalidXMLProcessorException;
	
	
	/* Decided to remove this service, as the client shouldn't get a processor-
	 * centric view, as he would still have to determine the correct method.
	 * Any comments?
	 * 
	 *  Gets the processor that will handle the given xml type. Returns the handler
	 * that would respond if the 'process(...)' method were invoked.
	 * @param type The Schema Type under consideration
	 * @return The handler which is best suited to handle this type, or null if there is none 
	 * 
	public XMLProcessor getProcessorForType( SchemaType  type);
	*/
	
	/** Exposes the underlying type system in use by the XML Processing Engine */
	public SchemaTypeRegistry getTypeSystem();
	
	
	/** A factory for creating XML processing engines. 
	 * TODO Currently tied to a single (basic) implementation
	 *  */
	public class Factory
	{
		/** Get an instance of a XML processing engine, which draws upon the schemas
		 * provided by the given schema locator.
		 * @return An instantiated (ready to use) processing engine.
		 * @throws Exception If the processing engine could not be created.
		 */
		public static XMLProcessingEngine getInstance( SchemaLocator locator ) throws Exception
		{
			// TODO Change this to more generic code
			return new BasicXMLProcessingEngine( locator );
		}
	}
}
