package za.co.solms.framework.ooxml.processing;

import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import nu.xom.Element;

/** An XML Processor is a pluggable handler for consuming XML messages,
 * and producing output based on the input message.
 *
 * Components that implement this interface form part of a pluggable message handling
 * framework, the key concept of which is Polymorphism on Message Arguments. Each 
 * implementation must annotate their "process" messages with one or both of
 * the @link XMLInstance or @link XMLType annotations.
 * 
 * @author dawidl
 * @version 0.1 (Nov 3, 2005)
 */
public interface XMLProcessor
{
	/** Processes the given input XML tree, producing output. Implementations should
	 * annotate this method with the @link XMLInstance and/or @link XMLType annotations,
	 * in order to allow the processing framework to select the correct implementations.
	 * From an XML perspective, these annotations are analagous to a java method like
	 * <code>public Object process( MyType type )</code>, with the benefit of having
	 * polymorphism on mesage arguments (which Java lacks). Non-annotated methods will
	 * not be seen as message processing services.
	 * 
	 * NOTE: This service specification ensures that a processor has at least one processing
	 * service. Any number of services with the correct input/output, e.g. 
	 * <code>public Element xxxx( Element input)</code> can be added to your implementation.
	 * If annotated, these will be called by the framework too (service name is irrelevant)
	 * 
	 * @param input The message to be processed. Will always be of the type (or subtype) 
	 * specified by the annotation.
	 * @return output. If no Exception si thrown, output must always be returned.
	 * @throws Exception to indicate failure. The XML framework will then respond to the client
	 * according to implementation-specific setup.
	 */
	public XMLMessage process( XMLMessage input ) throws Exception;
	
	
	/** Called exactly once (before any XML messages are handed to it for processing)
	 * in order to initialise.
	 * 
	 * TODO Instead of SchemaTypeRegistry, we need to provide a proper "context"
	 * 
	 * @param typeRegistry The type system in uses, which the processor can make use 
	 * of to look up types, etc.
	 */
	public void init( SchemaTypeRegistry typeRegistry) throws Exception;
}
