package za.co.solms.framework.ooxml.processing.tests;

import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import nu.xom.Builder;
import nu.xom.Document;
import org.apache.log4j.BasicConfigurator;
import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.processing.UnhandledMessageTypeException;
import za.co.solms.framework.ooxml.processing.XMLProcessingEngine;
import za.co.solms.framework.ooxml.processing.XMLProcessor;
import za.co.solms.framework.ooxml.processing.XMLType;
import za.co.solms.framework.ooxml.types.SchemaLocator;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import junit.framework.TestCase;

/**
 * 
 * TODO Add tests for: init() 
 *
 */
public class XMLProcessingEngineTest extends TestCase
{

	public static void main(String[] args)
	{
		junit.textui.TestRunner.run(XMLProcessingEngineTest.class);
	}
	
	public XMLProcessingEngineTest() throws Exception
	{
		// Logging
		//BasicConfigurator.configure();
	}

	protected void setUp() throws Exception
	{
		engine = XMLProcessingEngine.Factory.getInstance( new TestSchemaLocator() );
		processor = new TestProcessor();
	}


	/*
	 * Test method for 'za.co.solms.framework.impl.ooxml.basic.processing.BasicXMLProcessingEngine.addProcessor(XMLProcessor)'
	 */
	public void testAddProcessor() throws Exception
	{
		// Our dummy message
		Document a = xmlBuilder.build( getClass().getClassLoader().getResourceAsStream( TEST_PATH + "a.xml" ) );
		XMLMessage aMsg = new XMLMessage( a );
		
		// Send messae (expect error)
		boolean err = false;
		try
		{
			engine.process( aMsg );
		}
		catch (UnhandledMessageTypeException e)
		{
			err = true;
		}
		assertTrue( err );
		
		
		// Add Processor
		engine.addProcessor( processor );
		
		// Send message again (should not throw any errors now) - ignore response
		engine.process( aMsg );
	}

	/*
	 * Test method for 'za.co.solms.framework.impl.ooxml.basic.processing.BasicXMLProcessingEngine.removeProcessor(XMLProcessor)'
	 */
	public void testRemoveProcessor() throws Exception
	{
		// Call test that adds processor (and doesn't remove it afterwards)
		testAddProcessor();
		
		// Our dummy message
		Document a = xmlBuilder.build( getClass().getClassLoader().getResourceAsStream( TEST_PATH + "a.xml" ) );
		XMLMessage aMsg = new XMLMessage( a );
		
		// Make sure, if you like
		engine.process( aMsg );
		
		// Remove Processor
		engine.removeProcessor( processor );
		
		// Send Message, and expect error (other errors should filter through though)
		boolean err = false;
		try
		{
			engine.process( aMsg );
		}
		catch (UnhandledMessageTypeException e)
		{
			err = true;
		}
		assertTrue( err );
		
	}

	/*
	 * Test method for 'za.co.solms.framework.impl.ooxml.basic.processing.BasicXMLProcessingEngine.process(Element)'
	 */
	public void testProcess() throws Exception
	{
		// TEST POLYMORPHISM ("ON MESSAGE ARGUMENTS")
		// Send plain 'A' message
		TestProcessor aProcessor = new TestProcessor();
		engine.addProcessor( aProcessor );
		Document a = xmlBuilder.build( getClass().getClassLoader().getResourceAsStream( TEST_PATH + "a.xml" ) );
		XMLMessage aMsg = new XMLMessage( a );
		
		engine.process( aMsg );
		assertTrue( aProcessor.isAInvoked() );
		aProcessor.reset();
		
		// Send derived 'ADerived' message
		Document aDerived = xmlBuilder.build( getClass().getClassLoader().getResourceAsStream( TEST_PATH + "aDerived.xml" ) );
		XMLMessage aDerivedMsg = new XMLMessage( aDerived );
		engine.process( aDerivedMsg );
		assertTrue( aProcessor.isAInvoked() );
		aProcessor.reset();
		
		// Send derived 'B' (different namespace, different schema) message
		Document b = xmlBuilder.build( getClass().getClassLoader().getResourceAsStream( TEST_PATH + "bDerivedFromA.xml" ) );
		XMLMessage bMsg = new XMLMessage( b );
		engine.process( bMsg );
		assertTrue( aProcessor.isAInvoked() );
		aProcessor.reset();
		
		// TEST DELEGATION
		// Now, add a new processor for type 'B', and send a B message again. This time,
		// processor for 'A' should not have been invoked.
		TestProcessorForB bProcessor = new TestProcessorForB();
		engine.addProcessor( bProcessor );
		
		engine.process( bMsg );
		assertFalse( aProcessor.isAInvoked() );
		assertTrue( bProcessor.isBInvoked() );
		
		// For good measure remove it, and make sure it's handled more generically then
		engine.removeProcessor( bProcessor );
		aProcessor.reset();
		engine.process( bMsg );
		assertTrue( aProcessor.isAInvoked() );
		
		// Now, add the combined processor, and send A and B messages to it
		// Because it's added last, it should 'overwrite' the previous mappings.
		TestProcessorForAB abProcessor = new TestProcessorForAB();
		engine.addProcessor( abProcessor );
		
		engine.process( aMsg );
		engine.process( bMsg );
		
		assertTrue( abProcessor.isAInvoked() );
		assertTrue( abProcessor.isBInvoked() );
		
	}

	private XMLProcessingEngine engine;
	private TestProcessor processor;
	private Builder xmlBuilder = new Builder();
	private String TEST_PATH = "za/co/solms/framework/ooxml/processing/tests/";
	
	
	/** A schema locator to provide the testing schemas */
	private class TestSchemaLocator implements SchemaLocator
	{
		public TestSchemaLocator()
		{
			schemaURLs = new HashSet<URL>();
			schemaURLs.add( getClass().getClassLoader().getResource(SCHEMA_PATH + "a.xsd") );
			schemaURLs.add( getClass().getClassLoader().getResource(SCHEMA_PATH + "b.xsd") );
		}
		
		public Set<URL> getSchemaURLs()
		{
			return schemaURLs;
		}
		
		private Set<URL> schemaURLs;
		private String SCHEMA_PATH = TEST_PATH + "schemas/";
	}
	
	
	/** A test processor - it simply records wether it was called or not. Based
	 * on the a.xsd and b.xsd test schemas. */
	public class TestProcessor implements XMLProcessor
	{
		/** Processor for messages of type 'A' */
		@XMLType(name="A", nameSpace="urn:foo:a")
		public XMLMessage process(XMLMessage input) throws Exception
		{
			System.out.println("Processor (A) invoked with message: " + input );
			aInvoked = true;
			
			// Respond with request message
			return input;
		}
		
		/** Initialisation */
		public void init(SchemaTypeRegistry typeRegistry){}
		
		// State reporting and reset methods
		public boolean isAInvoked(){ return aInvoked; }
		public void reset(){ aInvoked = false; }
		private boolean aInvoked;
	}
	
	
	/** A test processor - it simply records wether it was called or not. Based
	 * on the a.xsd and b.xsd test schemas. */
	public class TestProcessorForB implements XMLProcessor
	{
		/** Processor for messages of type 'B' */
		@XMLType(name="BDerivedFromA", nameSpace="urn:foo:b")
		public XMLMessage process(XMLMessage input) throws Exception
		{
			System.out.println("Processor (B) invoked with message: " + input );
			bInvoked = true;
			
			// Respond with request message
			return input;
		}
		
		/** Initialisation */
		public void init(SchemaTypeRegistry typeRegistry){}
		
		// State reporting and reset methods
		public boolean isBInvoked(){ return bInvoked; }
		public void reset(){ bInvoked = false; }
		private boolean bInvoked;
	}
	
	/** A test processor - it simply records wether it was called or not. Based
	 * on the a.xsd and b.xsd test schemas. */
	public class TestProcessorForAB implements XMLProcessor
	{
		/** Processor for messages of type 'A' */
		@XMLType(name="A", nameSpace="urn:foo:a")
		public XMLMessage process(XMLMessage input) throws Exception
		{
			System.out.println("Processor (AB):A invoked with message: " + input );
			aInvoked = true;
			
			// Respond with request message
			return input;
		}
		
		/** Processor for messages of type 'B' */
		@XMLType(name="BDerivedFromA", nameSpace="urn:foo:b")
		public XMLMessage eatANiceLittleB(XMLMessage input) throws Exception
		{
			System.out.println("Processor (AB):B invoked with message: " + input );
			bInvoked = true;
			
			// Respond with request message
			return input;
		}
		
		/** Initialisation */
		public void init(SchemaTypeRegistry typeRegistry){}
		
		// State reporting and reset methods
		public boolean isAInvoked(){ return aInvoked; }
		public boolean isBInvoked(){ return bInvoked; }
		public void reset(){ aInvoked = false; bInvoked = false; }
		private boolean aInvoked, bInvoked;
	}
}
