package za.co.solms.framework.ooxml.soap.processing;

import za.co.solms.framework.impl.ooxml.soap.processing.BasicSOAPProcessor;
import za.co.solms.framework.ooxml.processing.InvalidXMLProcessorException;
import za.co.solms.framework.ooxml.processing.XMLProcessor;
import za.co.solms.framework.ooxml.soap.SOAPMessage;
import za.co.solms.framework.ooxml.types.SchemaLocator;

/** A component that processes SOAP messages, also called a SOAP Stack.
 * Treats SOAP messages in a pure, OO realm, and does not contain any
 * transport layer mappings / logic. Furthermore, no concession is made to
 * support non WS-I (Basic Profile 1.1) features, i.e. only document
 * literal style should be supproted.
 * 
 * Delegates processing to plugged in Processors. Completely responsible
 * for performing Exception -> Fault mapping, as well as creating the complete
 * response message. It is the transport layer's responsibility to
 * interrogate the SOAP message and apply any transport-specific semantics, like
 * HTTP Status Codes, etc.
 * 
 * A SOAP Processor is created using the enclosed factory.
 * 
 * TODO How is SOAP versioning handled? (currently 1.1)
 * TODO Support SOAP Headers (Add services to plug in 'code modules' for headers,
 * or, more elegantly, introduce a different type of annotation for XML Processors)
 * 
 * @see za.co.solms.framework.ooxml.processing.XMLProcessingEngine
 * @author dawidl
 * @version 0.1 (Nov 30, 2005)
 */
public interface SOAPProcessor
{

	/** Submits a SOAP message for processing, with the indicated response.
	 * If there is no response (i.e. this is a one-way (asynchronous) service,
	 * null is returned. Otherwise, the contents of the returned message must
	 * be sent to the client. SOAP Processors do not typically perform validation -
	 * this needs to have been done at a prior layer. There is also no guarantee
	 * that plugged-in processor create valid responses, and this is not checked for.
	 * 
	 * @param request
	 * @return The response to be sent to the client, or null if one-way
	 */
	public SOAPMessage process( SOAPMessage request );
	
	
	/** Adds an endpoint processor, which is immediately available for polymorphic
	 * XML message dispatch.
	 * @param processor
	 */
	public void addProcessor( XMLProcessor processor ) throws InvalidXMLProcessorException;
	
	
	/** Remove the specified endpoint processor. Any operations in progress, or operations
	 * queued internally (already dispatched to this processor) must be allowed 
	 * to complete normally.
	 * @param processor
	 */
	public void removeProcessor( XMLProcessor processor ) throws InvalidXMLProcessorException;
	
	
	/** Returns a locator which can provide the URL of the WSDL upon which this
	 * SOAP processor is based, 
	 */
	public WSDLLocator getWSDLLocator();
	
	
	/** Instructs the SOAP processor to use the specified factory to map
	 * Java exceptions (thrown anywhere during the processing pipeline) to
	 * SOAP fault messages. If thsi is not set, the SOAP processor implementation
	 * will use a default factory.
	 * @param factory The fatcory to use. If null, a default will be used.
	 */
	public void setFaultFactory( SOAPFaultFactory factory );
	
	
	/** A factory for producing SOAP Processors */
	public class Factory
	{
		/** Creates a SOAP Engine for the given WSDL and Schema catalog. 
		 * Any handlers (processors) may or may not have been plugged in, based
		 * on the implementation.
		 */
		public static SOAPProcessor getInstance(WSDLLocator wsdl, SchemaLocator schemas) throws Exception
		{
			// TODO Remove link to a specific implementation
			return new BasicSOAPProcessor(wsdl, schemas);
		}
	}
}
