package za.co.solms.framework.ooxml.types;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import za.co.solms.framework.ooxml.QualifiedName;

/** Represents a TYPE, found in a W3C XML Schema, abstractly (i.e. does not distinguish
 * between simple and complex type). A Type is qualified by it's name, and it's namespace (optional).
 * 
 * TODO This class is not Observable, i.e. does not notify listeners of state changes.
 * 
 * @author dawidl
 * @version 0.1 (Oct 17, 2005)
 *
 */
public class SchemaType implements Comparable
{
	/** Create with the given name (not allowed to be null). */
	public SchemaType( QualifiedName name ) throws IllegalArgumentException
	{
		if (name == null)
		{
			throw new IllegalArgumentException("Illegal (empty) type name");
		}
			
		this.name = name;
	}
	
	/** Create with the given name and namespace (does not extend anything). 
	 */
	public SchemaType(String nameSpaceName, String name) throws IllegalArgumentException
	{
		this( new QualifiedName( nameSpaceName, name) );
	}
	
	/** Create with the given name (no namespace, and does not extend anything). 
	 */
	public SchemaType(String name) throws IllegalArgumentException
	{
		this(null, name);
	}
	
	
	/** Create with the given name and namespace, and extending the given supertype. */
	public SchemaType(QualifiedName name, SchemaType superType) throws IllegalArgumentException
	{
		this( name );
		setSuperType( superType );
	}
	
	
	/**
	 * @return Returns the name of the type, as a qualified name.
	 */
	public QualifiedName getName()
	{
		return name;
	}
	
	
	/** Returns the supertype (if any) of this schema type */
	public SchemaType getSuperType()
	{
		return superType;
	}
	
	
	/** Test if this schema type is of the given type. This means either literally,
	 * or through abstraction (directly or indirectly). i.e. employee.isOfType(manager) = true
	 */
	public boolean isOfType( SchemaType type )
	{
		if (type == null)
			return false;
				
		// Move up the type tree and evaulate
		SchemaType t = this;
		while (t != null)
		{
			if (type.equals( t ))
			{
				return true;
			}
			t = t.getSuperType();
		}
		return false;
	}

	
	/** Indicates all the known subtypes. Subtypes are a Component of schema type, so a copy 
	 * of the set is returned. This returned copy is NOT THREADSAFE.
	 */
	public Set<SchemaType> getKnownSubTypes()
	{
		return new HashSet<SchemaType>( knownSubTypes );
	}
	
	
	/** Add a subtype to this type */
	public void addSubType(SchemaType type)
	{
		if (type != null)
		{
			// Ensure we can't set ourselves as subtype
			if (this.equals(type))
			{
				return;
			}
			// Add to set
			knownSubTypes.add( type );
		}
	}
	
	/** Remove a subtype from this type */
	public void removeSubType(SchemaType type)
	{
		knownSubTypes.remove( type );
	}
	
	/** Sets the supertype. Not meant for run-time use, only during construction.
	 * If the supertype is changed, however, it is correctly un-referenced bidirectionally.
	 * @param type
	 */
	protected void setSuperType( SchemaType type )
	{
		// Ensure we can't set ourselves as supertype
		if (this.equals(type))
		{
			return;
		}
		
		// First, unlink us from the current supertype if present
		if (superType != null && superType != type)
		{
			superType.removeSubType( this );
		}
		
		// Then, try to link to this supertype
		if (type != null)
		{
			type.addSubType( this );
		}
		
		this.superType = type;
	}
	
	
	/** Accesses the type registry in which this type is resident. There is
	 * no public way to set this - it is expected that SchemaTypeRegistries
	 * will provide their own mechanism to set it at loading time (e.g.
	 * subclassing)*/
	public SchemaTypeRegistry getTypeRegistry()
	{
		return typeRegistry;
	}
	
	
	/** Simple string representation. WARNING: Do not use for matching, use one of the other methods,
	 * such as getQualifiedName().
	 */
	public String toString() 
	{
		// Name
		String s = getName().toString();
				
		if (superType != null)
		{
			s += " (extends: " + superType.getName().toString() + ")";
		}
		
		int sSize = knownSubTypes.size(); 
		if (sSize > 0)
		{
			s += " (" + sSize + " descendant" + (sSize != 1 ? "s" : "") + ")";
		}
				
		return s;
	}
	
	
	/** Two SchemaType references are considered equal if they refer to the same
	 * instance, of the instances have identical fully qualified names.
	 */
	public boolean equals(Object other) 
	{
		// Instance check
		if (other == this)
		{
			return true;
		}
		
		// Qualified name
		try
		{
			SchemaType o = (SchemaType)other;
			return o.getName().equals( name );
		}
		catch (Exception e)
		{
			return false;
		}
	}
	
	
	/** Calculates the HashCode of the type. Uses java.lang.String's implementation on the
	 * fully qualified name. */
	public int hashCode()
	{
		// Delegate to String's implementation
		return getName().hashCode();
	}
	
	
	/** Compares to another abject, to determine natural ordering. Based on type name only */
	public int compareTo(Object other)
	{
		try
		{
			SchemaType o = (SchemaType)other;
			return o.getName().getName().compareTo( name.getName() );
		}
		catch (Exception e)
		{
			return 0;
		}
	}
	
	
	//@Composite
	private QualifiedName name;
	//@Associate
	protected SchemaType superType;
	//@Associate
	private Set<SchemaType> knownSubTypes = new CopyOnWriteArraySet<SchemaType>();
	//@Associate
	protected SchemaTypeRegistry typeRegistry;
}
