package za.co.solms.framework.ooxml.types;

import java.io.IOException;
import java.net.URL;

import za.co.solms.framework.impl.ooxml.basic.types.BasicSchemaTypeRegistry;
import za.co.solms.framework.ooxml.QualifiedName;
import za.co.solms.framework.ooxml.XMLElementReference;

/** A Schema Type Registry is responsible for containing metadata of, and relationships between
 * the types (simple and complex) found in a collection of schema documents. It exposes services
 * to query these relationships, typically used at a low level in an Object-Oriented
 * XML framework, to, for example, allow polymorphism and substitutability.
 * 
 * The type registry aims to provide a flexible, dynamic type system in a pure XML
 * solution, as opposed to XML-to-Java mapping schemes such as JAXB or XMLBeans (which require
 * code generation and classloading). This framework does not contain the types themselves, nor does
 * it attempt to convert the types from the W3C XML Schema realm to the Java realm (as per JAXB). It
 * merely contains meta-information about the types, such as their relationships.
 * 
 * NOTE: A schema type registry needs to inherently contain all the W3C XML Schema types without
 * the user having to manually add these.
 * 
 * This registry is considered a low-level component, and presents both a "user" and a "code"
 * view through the services it specifies. It will typically be incorporated as part of a higher-level
 * framework.
 * 
 * @author dawidl
 * @version 0.1 (Oct 17, 2005)
 *
 */
public interface SchemaTypeRegistry
{
	/** Gets a reference to the specified schema type.
	 */
	public SchemaType getTypeByName( QualifiedName typeName ) throws TypeNotFoundException;
	
	
	/** Gets the type for the given (root) element name, i.e. if the xml contains <n:fluffy> (which, expanded,
	 * might be {http://animals.org}fluffy) this service might tell you that it is 
	 * of type {http://animals.org}Dog. IMPORTANT: This only works for the root element of a document,
	 * because XML elements do not inherently contain their type (it can only be inferred from their context).
	 * @param elementName The name of the element
	 * @throws TypeNotFoundException If the root element does not exist in the registry
	 */
	public SchemaType getTypeFromRootElement( QualifiedName elementName ) throws TypeNotFoundException;
	
	
	/** Gets the root element (i.e. 'element' as defined by a schema) - from which
	 * type information etc can be obtained.
	 * 
	 * @param elementName The name of the element 
	 * @return A reference to the XMLElement
	 * @throws ElementNotFoundException If no such root element exists in the type registry.
	 */
	public XMLElementReference getRootElementByName( QualifiedName elementName ) throws ElementNotFoundException;
	
	
	/** Adds the given type to the registry. Duplicates are ignored. */
	public void addType( SchemaType type );
	
	
	/** Adds all types from the given schema to the registry. Must happen synchronously, i.e. 
	 * when the method returns, it must be assumed that all types (if any) have been added 
	 * to the registry, and are ready for immediate use. The schema may or may not be checked 
	 * for validity, depending on the implementation. Any indirectly referenced schemas are
	 * added automatically and synchronously (and firstly).*/
	public void addTypesFromSchema( URL url ) throws IOException, SchemaException;
	
	
	/** Removes the given type from the schema. It is un-referenced from all other types that have
	 * a relationship with this type, unless the type is being "used" (i.e. other types depend on it)
	 * in which case it cannot be removed.*/
	public void removeType( SchemaType type ) throws TypeInUseException, TypeNotFoundException;
	
	
	/** Links a valid (as per schema) root element to its type. This allows type lookups
	 * (though the getTypeFromRootElement(...) method). The type must already be in the registry.
	 * @param rootElementName The name of the root element
	 * @param type The type of which this root must be in roder to be considered valid as per XML Schema
	 * @throws TypeNotFoundException If the referenced type does not exist in the registry
	 * @throws SchemaConflictException If the specified root element si already assigned to another type
	 */
	public void assignRootElementToType( QualifiedName rootElementName, SchemaType type ) throws TypeNotFoundException, SchemaConflictException;
	
	
	/** Provides instances that implement the 'SchemaTypeRegistry' interface.
	 * TODO Not pluggable! Currently tied to a single implementation of 'SchemaTypeRegistry'
	 */
	public class Factory
	{
		/** Returns a new instance of a SchemaTypeRegistry. At this stage, always
		 * returns a new 'BasicSchemaTypeRegistry' (not a singleton).
		 * @return
		 */
		public static SchemaTypeRegistry getInstance() throws Exception
		{
			return new BasicSchemaTypeRegistry();
		}
	}
}