package za.co.solms.framework.ooxml.types.tests;

import java.io.IOException;
import java.net.URL;

import org.apache.log4j.BasicConfigurator;

import za.co.solms.framework.ooxml.QualifiedName;
import za.co.solms.framework.ooxml.XMLElementReference;
import za.co.solms.framework.ooxml.types.ElementNotFoundException;
import za.co.solms.framework.ooxml.types.SchemaConflictException;
import za.co.solms.framework.ooxml.types.SchemaException;
import za.co.solms.framework.ooxml.types.SchemaType;
import za.co.solms.framework.ooxml.types.SchemaTypeRegistry;
import za.co.solms.framework.ooxml.types.TypeNotFoundException;
import junit.framework.TestCase;

/** Test case for the schema types registry. Tests the CONTRACT (not a particular implementation thereof)
 * and it simply uses the standard factory to give us an instance of a type registry. 
 * 
 * @author dawidl
 * @version 0.1 (Oct 25, 2005)
 */
public class SchemaTypeRegistryTest extends TestCase
{

	public static void main(String[] args)
	{
		junit.textui.TestRunner.run(SchemaTypeRegistryTest.class);
	}
	
	public SchemaTypeRegistryTest()
	{
		// Commented out until we can figure out why output is duplicated x amount of test methods
		BasicConfigurator.configure();
	}
	
	@Override
	protected void setUp() throws Exception
	{
		registry = SchemaTypeRegistry.Factory.getInstance();
	}
	
	@Override
	protected void tearDown() throws Exception
	{
		registry = null;
	}

	
	/*
	 * Test method for 'za.co.solms.framework.ooxml.SchemaTypeRegistry.getTypeByName(QualifiedName)'
	 */
	public void testGetTypeByName() throws TypeNotFoundException
	{
		SchemaType t1 = new SchemaType("urn:foo", "Bar");
		registry.addType( t1 );
		SchemaType t2 = registry.getTypeByName( new QualifiedName("urn:foo", "Bar") );
		// Positive
		assertEquals( t1, t2 );
		assertSame( t1, t2 );
		// Negative
		boolean caughtError = false;
		try
		{
			registry.getTypeByName( new QualifiedName("urn:foo", "Koos") );
		}
		catch (TypeNotFoundException tnfe)
		{
			caughtError = true;
		}
		assertTrue( caughtError );
		
	}

	public void testRootElementOperations() throws SchemaConflictException, TypeNotFoundException, ElementNotFoundException
	{
		// Setup
		SchemaType t1 = new SchemaType( new QualifiedName("urn:animalplanet", "Animal"));
		SchemaType t2 = new SchemaType( new QualifiedName("urn:animalplanet","Canine"), t1);
		registry.addType( t1 );
		registry.addType( t2 );
		registry.assignRootElementToType( new QualifiedName("urn:animalplanet", "fluffy"), t2 );
		
		// Look up element
		XMLElementReference re2 = registry.getRootElementByName( new QualifiedName("urn:animalplanet", "fluffy") );

		// Look up type
		SchemaType t3 = registry.getTypeFromRootElement( new QualifiedName("urn:animalplanet", "fluffy") );
		
		// Elemen must refer to same type
		assertTrue( re2.getType() == t3 );
		
		// Test it
		assertEquals( t3, t2 );
		// Check OO Relationships
		assertTrue( t3.isOfType( t2 ) );
		assertTrue( t3.isOfType( t1 ) );
		// Check adding a root element, type of which is not yet in the registry (should fail)
		SchemaType t4 = new SchemaType( new QualifiedName("urn:vet.co.za","BullDog"), t2);
		boolean notFoundEx = false;
		try
		{
			registry.assignRootElementToType( new QualifiedName("urn:animalplanet", "butch"), t4 );
		} catch (TypeNotFoundException tnfe)
		{
			notFoundEx = true;
		}
		assertTrue( notFoundEx );
		// Now add the type, and then do it again (should succeed this time)
		registry.addType( t4 );
		registry.assignRootElementToType( new QualifiedName("urn:animalplanet", "butch"), t4 );
		// Now, try to add an already-added root but with a different type (should fail)
		notFoundEx = false;
		try
		{
			registry.assignRootElementToType( new QualifiedName("urn:animalplanet", "fluffy"), t1 );
		} catch (SchemaConflictException sce)
		{
			notFoundEx = true;
		}
		assertTrue( notFoundEx );
	}

	/*
	 * Test method for 'za.co.solms.framework.ooxml.SchemaTypeRegistry.addType(SchemaType)'
	 */
	public void testAddType() throws TypeNotFoundException
	{
		// Get a type (should not be there) - then add it, and get it again
		boolean exception = false;
		try
		{
			registry.getTypeByName( new QualifiedName("urn:foo", "Bar"));
		}
		catch (TypeNotFoundException e)
		{
			exception = true;
		}
		assertTrue( exception );
		SchemaType t1 = new SchemaType("urn:cars","Jaguar");
		registry.addType( t1 );
		// Same instance of name
		SchemaType t2 = registry.getTypeByName( t1.getName() );
		// Equivalent name
		SchemaType t3 = registry.getTypeByName( new QualifiedName("urn:cars","Jaguar") );
		// Equality
		assertEquals(t1, t2);
		assertEquals(t2, t3);
		// Should be same instance
		assertTrue( t1 == t2 );
		assertTrue( t2 == t3 );
	}

	/*
	 * Test method for 'za.co.solms.framework.ooxml.SchemaTypeRegistry.addTypesFromSchema(URL)'
	 */
	public void testAddTypesFromSchema() throws IOException, SchemaException, TypeNotFoundException
	{
		// Get the URL of the test schema (Foo Simple)
		URL schema = getClass().getClassLoader().getResource( "za/co/solms/framework/ooxml/types/tests/schemas/foo-simple.xsd" );
		
		// Add it
		registry.addTypesFromSchema( schema );
		
		// Check that the types are exactly as per schema
		String ns = "http://solms.co.za/ooxml/types/tests";
		assertNotNull( registry.getTypeByName( new QualifiedName(ns, "Foo") ) );
		assertNotNull( registry.getTypeByName( new QualifiedName(ns, "Bar") ) );
		// This is supposed to fail (no such type, this is an element)
		boolean notfound = false;
		try
		{
			registry.getTypeByName( new QualifiedName(ns, "foo") );
		}
		catch (TypeNotFoundException tnfe)
		{
			notfound = true;
		}
		assertTrue( notfound );
		
		// Check the root elements
		assertEquals(
			registry.getTypeFromRootElement( new QualifiedName(ns, "foo") ), 
			registry.getTypeByName( new QualifiedName(ns, "Foo") )
		);
		assertEquals(
			registry.getTypeFromRootElement( new QualifiedName(ns, "bar") ), 
			registry.getTypeByName( new QualifiedName(ns, "Bar") )
		);
		
		/* Check the <foo> is a Foo */
		assertEquals(
			registry.getTypeFromRootElement( new QualifiedName(ns, "foo") ),
			registry.getTypeByName( new QualifiedName(ns, "Foo") ) );
		
		/* Check basic polymorphism from root element, i.e. <foo> must be able to indicate
		 * that ExpandedFoo is a legal, substitutable type (known subtypes).
		 */
		assertTrue( registry.getTypeFromRootElement( new QualifiedName(ns, "foo") ).
		getKnownSubTypes().contains( registry.getTypeByName( new QualifiedName(ns, "ExpandedFoo") ) ) );
	}
	
	/*
	 * Test method for 'za.co.solms.framework.ooxml.SchemaTypeRegistry.addTypesFromSchema(URL)'
	 */
	public void testAddTypesFromSchemaWithImports() throws IOException, SchemaException, TypeNotFoundException
	{
		// MULTIPLE SCHEMAS
		// Get the URL of the test schema (Foo Simple)
		URL schema = getClass().getClassLoader().getResource( "za/co/solms/framework/ooxml/types/tests/schemas/import-simple.xsd" );
		
		// Add it (it's imports will be added as well)
		registry.addTypesFromSchema( schema );
		
		String ns = "http://solms.co.za/ooxml/types/tests";
		String nsIm = "http://solms.co.za/ooxml/types/tests/fleem";
		// Check basic polymorphism from root element, i.e. <foo> must be able to indicate
		// that ExpandedFoo is a legal, substitutable type (known subtypes).
		assertTrue( registry.getTypeFromRootElement( new QualifiedName(ns, "foo") ).
		getKnownSubTypes().contains( registry.getTypeByName( new QualifiedName(nsIm, "Fleem") ) ) );
	}

	/*
	 * Test method for 'za.co.solms.framework.ooxml.SchemaTypeRegistry.removeType(SchemaType)'
	 */
	public void testRemoveType()
	{
		// TODO Add tests
		// try to remove dependent types
	}
	
	/* Tests the built-in (schema namespace) types, such as String, etc... */
	public void testBuiltInTypes()
	{
		/* Test for existence of basic types (string, etc) */
		boolean failed = false;
		SchemaType t1 = null;
		try
		{
			t1 = registry.getTypeByName( new QualifiedName("http://www.w3.org/2001/XMLSchema", "anySimpleType") );
		}
		catch (Exception e)
		{
			failed = true;
		}
		assertTrue( !failed && t1 != null );
		
		failed = false;
		SchemaType t2 = null;
		try
		{
			t2 = registry.getTypeByName( new QualifiedName("http://www.w3.org/2001/XMLSchema", "string") );
		}
		catch (Exception e)
		{
			failed = true;
		}
		assertTrue( !failed && t2 != null );
		
		/* Test built-in types extending each other */
		// (string must extend anySimpleType)
		assertEquals( t2.getSuperType(), t1 );
		
		
		/* Now test for complex types */
		failed = false;
		SchemaType ct1 = null;
		try
		{
			ct1 = registry.getTypeByName( new QualifiedName("http://www.w3.org/2001/XMLSchema", "annotated") );
		}
		catch (Exception e)
		{
			failed = true;
		}
		assertTrue( !failed && ct1 != null );
		
		failed = false;
		SchemaType ct2 = null;
		try
		{
			ct2 = registry.getTypeByName( new QualifiedName("http://www.w3.org/2001/XMLSchema", "element") );
		}
		catch (Exception e)
		{
			failed = true;
		}
		assertTrue( !failed && ct2 != null );
		
		/* Test built-in types extending each other */
		// (element must extend annotated)
		assertEquals( ct2.getSuperType(), ct1 );
	}
	
	private SchemaTypeRegistry registry;
}
